from __future__ import annotations

import json

import aiohttp

from .. import exceptions
from .. import sync
from .config import CollectionConfig
from .discover import DiscoverResult
from .discover import collections_for_pair
from .discover import storage_instance_from_config
from .utils import JobFailed
from .utils import cli_logger
from .utils import get_status_name
from .utils import handle_cli_error
from .utils import load_status
from .utils import manage_sync_status
from .utils import save_status


async def prepare_pair(pair_name, collections, config, *, connector):
    pair = config.get_pair(pair_name)

    all_collections = dict(
        await collections_for_pair(
            status_path=config.general["status_path"],
            pair=pair,
            connector=connector,
        )
    )

    for collection_name in collections or all_collections:
        try:
            config_a, config_b = all_collections[collection_name]
        except KeyError:
            raise exceptions.UserError(
                "Pair {}: Collection {} not found. These are the "
                "configured collections:\n{}".format(
                    pair_name, json.dumps(collection_name), list(all_collections)
                )
            )

        collection = CollectionConfig(pair, collection_name, config_a, config_b)
        yield collection, config.general


async def sync_collection(
    collection,
    general,
    force_delete,
    *,
    connector: aiohttp.TCPConnector,
):
    pair = collection.pair
    status_name = get_status_name(pair.name, collection.name)

    try:
        cli_logger.info(f"Syncing {status_name}")

        a = await storage_instance_from_config(collection.config_a, connector=connector)
        b = await storage_instance_from_config(collection.config_b, connector=connector)

        sync_failed = False

        def error_callback(e):
            nonlocal sync_failed
            sync_failed = True
            handle_cli_error(status_name, e)

        with manage_sync_status(
            general["status_path"], pair.name, collection.name
        ) as status:
            await sync.sync(
                a,
                b,
                status,
                conflict_resolution=pair.conflict_resolution,
                force_delete=force_delete,
                error_callback=error_callback,
                partial_sync=pair.partial_sync,
            )

        if sync_failed:
            raise JobFailed
    except JobFailed:
        raise
    except BaseException:
        handle_cli_error(status_name)
        raise JobFailed


async def discover_collections(pair, **kwargs):
    rv = await collections_for_pair(pair=pair, **kwargs)
    collections = [c for c, (a, b) in rv]
    if collections == [None]:
        collections = None
    cli_logger.info(f"Saved for {pair.name}: collections = {json.dumps(collections)}")


async def repair_collection(
    config,
    collection,
    repair_unsafe_uid,
    *,
    connector: aiohttp.TCPConnector,
):
    from ..repair import repair_storage

    storage_name, collection = collection, None
    if "/" in storage_name:
        storage_name, collection = storage_name.split("/")

    config = config.get_storage_args(storage_name)
    # If storage type has a slash, ignore it and anything after it.
    storage_type = config["type"].split("/")[0]

    if collection is not None:
        cli_logger.info("Discovering collections (skipping cache).")
        get_discovered = DiscoverResult(config, connector=connector)
        discovered = await get_discovered.get_self()
        for config in discovered.values():
            if config["collection"] == collection:
                break
        else:
            raise exceptions.UserError(
                f"Couldn't find collection {collection} for storage {storage_name}."
            )

    config["type"] = storage_type
    storage = await storage_instance_from_config(config, connector=connector)

    cli_logger.info(f"Repairing {storage_name}/{collection}")
    cli_logger.warning("Make sure no other program is talking to the server.")
    await repair_storage(storage, repair_unsafe_uid=repair_unsafe_uid)


async def metasync_collection(collection, general, *, connector: aiohttp.TCPConnector):
    from ..metasync import metasync

    pair = collection.pair
    status_name = get_status_name(pair.name, collection.name)

    try:
        cli_logger.info(f"Metasyncing {status_name}")

        status = load_status(
            general["status_path"],
            pair.name,
            collection.name,
            data_type="metadata",
        )

        a = await storage_instance_from_config(collection.config_a, connector=connector)
        b = await storage_instance_from_config(collection.config_b, connector=connector)

        await metasync(
            a,
            b,
            status,
            conflict_resolution=pair.conflict_resolution,
            keys=pair.metadata,
        )
    except BaseException:
        handle_cli_error(status_name)
        raise JobFailed

    save_status(
        base_path=general["status_path"],
        pair=pair.name,
        data_type="metadata",
        data=status,
        collection=collection.name,
    )



"""
Vdirsyncer synchronizes calendars and contacts.
"""

from __future__ import annotations

PROJECT_HOME = "https://github.com/pimutils/vdirsyncer"
BUGTRACKER_HOME = PROJECT_HOME + "/issues"
DOCS_HOME = "https://vdirsyncer.pimutils.org/en/stable"

try:
    from .version import version as __version__  # noqa
except ImportError:  # pragma: no cover
    raise ImportError(
        "Failed to find (autogenerated) version.py. "
        "This might be because you are installing from GitHub's tarballs, "
        "use the PyPI ones."
    )


def _check_python_version():
    import sys

    if sys.version_info < (3, 8, 0):  # noqa: UP036
        print("vdirsyncer requires at least Python 3.8.")
        sys.exit(1)


_check_python_version()
del _check_python_version



"""
The `sync` function in `vdirsyncer.sync` can be called on two instances of
`Storage` to synchronize them. Apart from the defined errors, this is the only
public API of this module.

The algorithm is based on the blogpost "How OfflineIMAP works" by Edward Z.
Yang: http://blog.ezyang.com/2012/08/how-offlineimap-works/

Some modifications to it are explained in
https://unterwaditzer.net/2016/sync-algorithm.html
"""

from __future__ import annotations

import contextlib
import itertools
import logging

from vdirsyncer.storage.base import Storage
from vdirsyncer.vobject import Item

from ..exceptions import UserError
from ..utils import uniq
from .exceptions import BothReadOnly
from .exceptions import IdentAlreadyExists
from .exceptions import PartialSync
from .exceptions import StorageEmpty
from .exceptions import SyncConflict
from .status import ItemMetadata
from .status import SqliteStatus
from .status import SubStatus

sync_logger = logging.getLogger(__name__)


class _StorageInfo:
    """A wrapper class that holds prefetched items, the status and other
    things."""

    def __init__(self, storage: Storage, status: SubStatus):
        self.storage = storage
        self.status = status
        self._item_cache = {}  # type: ignore[var-annotated]

    async def prepare_new_status(self) -> bool:
        storage_nonempty = False
        prefetch = []

        def _store_props(ident: str, props: ItemMetadata) -> None:
            try:
                self.status.insert_ident(ident, props)
            except IdentAlreadyExists as e:
                raise e.to_ident_conflict(self.storage)

        async for href, etag in self.storage.list():  # type: ignore[attr-defined]
            storage_nonempty = True
            ident, meta = self.status.get_by_href(href)

            if meta is None or meta.href != href or meta.etag != etag:
                # Either the item is completely new, or updated
                # In both cases we should prefetch
                prefetch.append(href)
            else:
                # Metadata is completely identical
                _store_props(ident, meta)

        # Prefetch items
        if prefetch:
            async for href, item, etag in self.storage.get_multi(prefetch):
                _store_props(
                    item.ident,
                    ItemMetadata(href=href, hash=item.hash, etag=etag),
                )
                self.set_item_cache(item.ident, item)

        return storage_nonempty

    def is_changed(self, ident: str) -> bool:
        old_meta = self.status.get(ident)
        if old_meta is None:  # new item
            return True

        new_meta = self.status.get_new(ident)

        return (
            new_meta.etag != old_meta.etag  # etag changed
            # item actually changed
            and (old_meta.hash is None or new_meta.hash != old_meta.hash)
        )

    def set_item_cache(self, ident, item) -> None:
        actual_hash = self.status.get_new(ident).hash
        assert actual_hash == item.hash
        self._item_cache[ident] = item

    def get_item_cache(self, ident: str) -> Item:
        return self._item_cache[ident]


async def sync(
    storage_a: Storage,
    storage_b: Storage,
    status: SqliteStatus,
    conflict_resolution=None,
    force_delete=False,
    error_callback=None,
    partial_sync="revert",
) -> None:
    """Synchronizes two storages.

    :param storage_a: The first storage
    :param storage_b: The second storage
    :param status: {ident: (href_a, etag_a, href_b, etag_b)}
        metadata about the two storages for detection of changes. Will be
        modified by the function and should be passed to it at the next sync.
        If this is the first sync, an empty dictionary should be provided.
    :param conflict_resolution: A function that, given two conflicting item
        versions A and B, returns a new item with conflicts resolved. The UID
        must be the same. The strings `"a wins"` and `"b wins"` are also
        accepted to mean that side's version will always be taken. If none
        is provided, the sync function will raise :py:exc:`SyncConflict`.
    :param force_delete: When one storage got completely emptied between two
        syncs, :py:exc:`StorageEmpty` is raised for
        safety. Setting this parameter to ``True`` disables this safety
        measure.
    :param error_callback: Instead of raising errors when executing actions,
        call the given function with an `Exception` as the only argument.
    :param partial_sync: What to do when doing sync actions on read-only
        storages.

        - ``error``: Raise an error.
        - ``ignore``: Those actions are simply skipped.
        - ``revert`` (default): Revert changes on other side.
    """
    if storage_a.read_only and storage_b.read_only:
        raise BothReadOnly

    if conflict_resolution == "a wins":
        conflict_resolution = lambda a, b: a  # noqa: E731
    elif conflict_resolution == "b wins":
        conflict_resolution = lambda a, b: b  # noqa: E731

    status_nonempty = bool(next(status.iter_old(), None))

    with status.transaction():
        a_info = _StorageInfo(storage_a, SubStatus(status, "a"))
        b_info = _StorageInfo(storage_b, SubStatus(status, "b"))

        a_nonempty = await a_info.prepare_new_status()
        b_nonempty = await b_info.prepare_new_status()

        if status_nonempty and not force_delete:
            if a_nonempty and not b_nonempty:
                raise StorageEmpty(empty_storage=storage_b)
            elif not a_nonempty and b_nonempty:
                raise StorageEmpty(empty_storage=storage_a)

        actions = list(_get_actions(a_info, b_info))

        async with storage_a.at_once(), storage_b.at_once():
            for action in actions:
                try:
                    await action.run(a_info, b_info, conflict_resolution, partial_sync)
                except Exception as e:
                    if error_callback:
                        error_callback(e)
                    else:
                        raise


class Action:
    async def _run_impl(self, a, b):  # pragma: no cover
        raise NotImplementedError

    async def run(self, a, b, conflict_resolution, partial_sync):
        with self.auto_rollback(a, b):
            if self.dest.storage.read_only:
                if partial_sync == "error":
                    raise PartialSync(self.dest.storage)
                elif partial_sync == "ignore":
                    self.rollback(a, b)
                    return
                else:
                    assert partial_sync == "revert"

            await self._run_impl(a, b)

    @contextlib.contextmanager
    def auto_rollback(self, a, b):
        try:
            yield
        except BaseException as e:
            self.rollback(a, b)
            raise e

    def rollback(self, a, b):
        a.status.parent.rollback(self.ident)


class Upload(Action):
    def __init__(self, item, dest):
        self.item = item
        self.ident = item.ident
        self.dest = dest

    async def _run_impl(self, a, b):
        if self.dest.storage.read_only:
            href = etag = None
        else:
            sync_logger.info(
                f"Copying (uploading) item {self.ident} to {self.dest.storage}"
            )
            href, etag = await self.dest.storage.upload(self.item)
            assert href is not None

        self.dest.status.insert_ident(
            self.ident, ItemMetadata(href=href, hash=self.item.hash, etag=etag)
        )


class Update(Action):
    def __init__(self, item, dest):
        self.item = item
        self.ident = item.ident
        self.dest = dest

    async def _run_impl(self, a, b):
        if self.dest.storage.read_only:
            meta = ItemMetadata(hash=self.item.hash)
        else:
            sync_logger.info(
                f"Copying (updating) item {self.ident} to {self.dest.storage}"
            )
            meta = self.dest.status.get_new(self.ident)
            meta.etag = await self.dest.storage.update(meta.href, self.item, meta.etag)

        self.dest.status.update_ident(self.ident, meta)


class Delete(Action):
    def __init__(self, ident, dest):
        self.ident = ident
        self.dest = dest

    async def _run_impl(self, a, b):
        meta = self.dest.status.get_new(self.ident)
        if self.dest.storage.read_only or self.dest.storage.no_delete:
            sync_logger.debug(
                f"Skipping deletion of item {self.ident} from {self.dest.storage}"
            )
        else:
            sync_logger.info(f"Deleting item {self.ident} from {self.dest.storage}")
            await self.dest.storage.delete(meta.href, meta.etag)

        self.dest.status.remove_ident(self.ident)


class ResolveConflict(Action):
    def __init__(self, ident):
        self.ident = ident

    async def run(self, a, b, conflict_resolution, partial_sync):
        with self.auto_rollback(a, b):
            sync_logger.info(f"Doing conflict resolution for item {self.ident}...")

            meta_a = a.status.get_new(self.ident)
            meta_b = b.status.get_new(self.ident)

            if meta_a.hash == meta_b.hash:
                sync_logger.info("...same content on both sides.")
            elif conflict_resolution is None:
                raise SyncConflict(
                    ident=self.ident, href_a=meta_a.href, href_b=meta_b.href
                )
            elif callable(conflict_resolution):
                item_a = a.get_item_cache(self.ident)
                item_b = b.get_item_cache(self.ident)
                new_item = conflict_resolution(item_a, item_b)
                if new_item.hash != meta_a.hash:
                    await Update(new_item, a).run(
                        a,
                        b,
                        conflict_resolution,
                        partial_sync,
                    )
                if new_item.hash != meta_b.hash:
                    await Update(new_item, b).run(
                        a,
                        b,
                        conflict_resolution,
                        partial_sync,
                    )
            else:
                raise UserError(
                    f"Invalid conflict resolution mode: {conflict_resolution!r}"
                )


def _get_actions(a_info: _StorageInfo, b_info: _StorageInfo):
    for ident in uniq(
        itertools.chain(
            a_info.status.parent.iter_new(), a_info.status.parent.iter_old()
        )
    ):
        a = a_info.status.get_new(ident)
        b = b_info.status.get_new(ident)

        if a and b:
            a_changed = a_info.is_changed(ident)
            b_changed = b_info.is_changed(ident)
            if a_changed and b_changed:
                # item was modified on both sides
                # OR: missing status
                yield ResolveConflict(ident)
            elif a_changed and not b_changed:
                # item was only modified in a
                yield Update(a_info.get_item_cache(ident), b_info)
            elif not a_changed and b_changed:
                # item was only modified in b
                yield Update(b_info.get_item_cache(ident), a_info)
        elif a and not b:
            if a_info.is_changed(ident):
                # was deleted from b but modified on a
                # OR: new item was created in a
                yield Upload(a_info.get_item_cache(ident), b_info)
            else:
                # was deleted from b and not modified on a
                yield Delete(ident, a_info)
        elif not a and b:
            if b_info.is_changed(ident):
                # was deleted from a but modified on b
                # OR: new item was created in b
                yield Upload(b_info.get_item_cache(ident), a_info)
            else:
                # was deleted from a and not changed on b
                yield Delete(ident, b_info)



from __future__ import annotations

import collections
import contextlib
import functools
import glob
import logging
import os
from typing import Iterable

from .. import exceptions
from ..utils import atomic_write
from ..utils import checkfile
from ..utils import expand_path
from ..utils import get_etag_from_file
from ..utils import uniq
from ..vobject import Item
from ..vobject import join_collection
from ..vobject import split_collection
from .base import Storage

logger = logging.getLogger(__name__)


def _writing_op(f):
    @functools.wraps(f)
    async def inner(self, *args, **kwargs):
        if self._items is None or not self._at_once:
            async for _ in self.list():
                pass
        assert self._items is not None
        rv = await f(self, *args, **kwargs)
        if not self._at_once:
            self._write()
        return rv

    return inner


class SingleFileStorage(Storage):
    storage_name = "singlefile"
    _repr_attributes = ["path"]

    _write_mode = "wb"
    _append_mode = "ab"
    _read_mode = "rb"

    _items = None
    _last_etag = None

    def __init__(self, path, encoding="utf-8", **kwargs):
        super().__init__(**kwargs)
        path = os.path.abspath(expand_path(path))
        checkfile(path, create=False)

        self.path = path
        self.encoding = encoding
        self._at_once = False

    @classmethod
    async def discover(cls, path, **kwargs):
        if kwargs.pop("collection", None) is not None:
            raise TypeError("collection argument must not be given.")

        path = os.path.abspath(expand_path(path))
        try:
            path_glob = path % "*"
        except TypeError:
            # If not exactly one '%s' is present, we cannot discover
            # collections because we wouldn't know which name to assign.
            raise NotImplementedError

        placeholder_pos = path.index("%s")

        for subpath in glob.iglob(path_glob):
            if os.path.isfile(subpath):
                args = dict(kwargs)
                args["path"] = subpath

                collection_end = (
                    placeholder_pos + 2 + len(subpath) - len(path)  # length of '%s'
                )
                collection = subpath[placeholder_pos:collection_end]
                args["collection"] = collection

                yield args

    @classmethod
    async def create_collection(cls, collection, **kwargs):
        path = os.path.abspath(expand_path(kwargs["path"]))

        if collection is not None:
            try:
                path = path % (collection,)
            except TypeError:
                raise ValueError(
                    "Exactly one %s required in path if collection is not null."
                )

        checkfile(path, create=True)
        kwargs["path"] = path
        kwargs["collection"] = collection
        return kwargs

    async def list(self):
        self._items = collections.OrderedDict()

        try:
            self._last_etag = get_etag_from_file(self.path)
            with open(self.path, self._read_mode) as f:
                text = f.read().decode(self.encoding)
        except OSError as e:
            import errno

            if e.errno != errno.ENOENT:  # file not found
                raise OSError(e)
            text = None

        if text:
            for item in split_collection(text):
                item = Item(item)
                etag = item.hash
                href = item.ident
                self._items[href] = item, etag

                yield href, etag

    async def get(self, href):
        if self._items is None or not self._at_once:
            async for _ in self.list():
                pass

        try:
            return self._items[href]
        except KeyError:
            raise exceptions.NotFoundError(href)

    async def get_multi(self, hrefs: Iterable[str]):
        async with self.at_once():
            for href in uniq(hrefs):
                item, etag = await self.get(href)
                yield href, item, etag

    @_writing_op
    async def upload(self, item):
        href = item.ident
        if href in self._items:
            raise exceptions.AlreadyExistingError(existing_href=href)

        self._items[href] = item, item.hash
        return href, item.hash

    @_writing_op
    async def update(self, href, item, etag):
        if href not in self._items:
            raise exceptions.NotFoundError(href)

        _, actual_etag = self._items[href]
        if etag != actual_etag:
            raise exceptions.WrongEtagError(etag, actual_etag)

        self._items[href] = item, item.hash
        return item.hash

    @_writing_op
    async def delete(self, href, etag):
        if href not in self._items:
            raise exceptions.NotFoundError(href)

        _, actual_etag = self._items[href]
        if etag != actual_etag:
            raise exceptions.WrongEtagError(etag, actual_etag)

        del self._items[href]

    def _write(self):
        if self._last_etag is not None and self._last_etag != get_etag_from_file(
            self.path
        ):
            raise exceptions.PreconditionFailed(
                f"Some other program modified the file {self.path!r}. Re-run the "
                "synchronization and make sure absolutely no other program is "
                "writing into the same file."
            )
        text = join_collection(item.raw for item, etag in self._items.values())
        try:
            with atomic_write(self.path, mode="wb", overwrite=True) as f:
                f.write(text.encode(self.encoding))
        finally:
            self._items = None
            self._last_etag = None

    @contextlib.asynccontextmanager
    async def at_once(self):
        async for _ in self.list():
            pass
        self._at_once = True
        try:
            yield self
            self._write()
        finally:
            self._at_once = False



from __future__ import annotations

import errno
import logging
import os
import subprocess

from .. import exceptions
from ..utils import atomic_write
from ..utils import checkdir
from ..utils import expand_path
from ..utils import generate_href
from ..utils import get_etag_from_file
from ..vobject import Item
from .base import Storage
from .base import normalize_meta_value

logger = logging.getLogger(__name__)


class FilesystemStorage(Storage):
    storage_name = "filesystem"
    _repr_attributes = ["path"]

    def __init__(
        self,
        path,
        fileext,
        encoding="utf-8",
        post_hook=None,
        pre_deletion_hook=None,
        fileignoreext=".tmp",
        **kwargs,
    ):
        super().__init__(**kwargs)
        path = expand_path(path)
        checkdir(path, create=False)
        self.path = path
        self.encoding = encoding
        self.fileext = fileext
        self.fileignoreext = fileignoreext
        self.post_hook = post_hook
        self.pre_deletion_hook = pre_deletion_hook

    @classmethod
    async def discover(cls, path, **kwargs):
        if kwargs.pop("collection", None) is not None:
            raise TypeError("collection argument must not be given.")
        path = expand_path(path)
        try:
            collections = os.listdir(path)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise
        else:
            for collection in collections:
                collection_path = os.path.join(path, collection)
                if not cls._validate_collection(collection_path):
                    continue
                args = dict(collection=collection, path=collection_path, **kwargs)
                yield args

    @classmethod
    def _validate_collection(cls, path):
        if not os.path.isdir(path):
            return False
        if os.path.basename(path).startswith("."):
            return False
        return True

    @classmethod
    async def create_collection(cls, collection, **kwargs):
        kwargs = dict(kwargs)
        path = kwargs["path"]

        if collection is not None:
            path = os.path.join(path, collection)

        checkdir(expand_path(path), create=True)

        kwargs["path"] = path
        kwargs["collection"] = collection
        return kwargs

    def _get_filepath(self, href):
        return os.path.join(self.path, href)

    def _get_href(self, ident):
        return generate_href(ident) + self.fileext

    async def list(self):
        for fname in os.listdir(self.path):
            fpath = os.path.join(self.path, fname)
            if (
                os.path.isfile(fpath)
                and fname.endswith(self.fileext)
                and (not fname.endswith(self.fileignoreext))
            ):
                yield fname, get_etag_from_file(fpath)

    async def get(self, href):
        fpath = self._get_filepath(href)
        try:
            with open(fpath, "rb") as f:
                return (Item(f.read().decode(self.encoding)), get_etag_from_file(fpath))
        except OSError as e:
            if e.errno == errno.ENOENT:
                raise exceptions.NotFoundError(href)
            else:
                raise

    async def upload(self, item):
        if not isinstance(item.raw, str):
            raise TypeError("item.raw must be a unicode string.")

        try:
            href = self._get_href(item.ident)
            fpath, etag = self._upload_impl(item, href)
        except OSError as e:
            if e.errno in (errno.ENAMETOOLONG, errno.ENOENT):  # Unix  # Windows
                logger.debug("UID as filename rejected, trying with random one.")
                # random href instead of UID-based
                href = self._get_href(None)
                fpath, etag = self._upload_impl(item, href)
            else:
                raise

        if self.post_hook:
            self._run_post_hook(fpath)
        return href, etag

    def _upload_impl(self, item, href):
        fpath = self._get_filepath(href)
        try:
            with atomic_write(fpath, mode="wb", overwrite=False) as f:
                f.write(item.raw.encode(self.encoding))
                return fpath, get_etag_from_file(f)
        except OSError as e:
            if e.errno == errno.EEXIST:
                raise exceptions.AlreadyExistingError(existing_href=href)
            else:
                raise

    async def update(self, href, item, etag):
        fpath = self._get_filepath(href)
        if not os.path.exists(fpath):
            raise exceptions.NotFoundError(item.uid)
        actual_etag = get_etag_from_file(fpath)
        if etag != actual_etag:
            raise exceptions.WrongEtagError(etag, actual_etag)

        if not isinstance(item.raw, str):
            raise TypeError("item.raw must be a unicode string.")

        with atomic_write(fpath, mode="wb", overwrite=True) as f:
            f.write(item.raw.encode(self.encoding))
            etag = get_etag_from_file(f)

        if self.post_hook:
            self._run_post_hook(fpath)
        return etag

    async def delete(self, href, etag):
        fpath = self._get_filepath(href)
        if not os.path.isfile(fpath):
            raise exceptions.NotFoundError(href)
        actual_etag = get_etag_from_file(fpath)
        if etag != actual_etag:
            raise exceptions.WrongEtagError(etag, actual_etag)
        if self.pre_deletion_hook:
            self._run_pre_deletion_hook(fpath)

        os.remove(fpath)

    def _run_post_hook(self, fpath):
        logger.info(f"Calling post_hook={self.post_hook} with argument={fpath}")
        try:
            subprocess.call([self.post_hook, fpath])
        except OSError as e:
            logger.warning(f"Error executing external hook: {str(e)}")

    def _run_pre_deletion_hook(self, fpath):
        logger.info(
            f"Calling pre_deletion_hook={self.pre_deletion_hook} with argument={fpath}"
        )
        try:
            subprocess.call([self.pre_deletion_hook, fpath])
        except OSError as e:
            logger.warning(f"Error executing external hook: {str(e)}")

    async def get_meta(self, key):
        fpath = os.path.join(self.path, key)
        try:
            with open(fpath, "rb") as f:
                return normalize_meta_value(f.read().decode(self.encoding))
        except OSError as e:
            if e.errno == errno.ENOENT:
                return None
            else:
                raise

    async def set_meta(self, key, value):
        value = normalize_meta_value(value)

        fpath = os.path.join(self.path, key)
        if value is None:
            try:
                os.remove(fpath)
            except OSError:
                pass
        else:
            with atomic_write(fpath, mode="wb", overwrite=True) as f:
                f.write(value.encode(self.encoding))



from __future__ import annotations

import logging
import os
import platform
import re
from abc import ABC
from abc import abstractmethod
from base64 import b64encode
from ssl import create_default_context

import aiohttp
import requests.auth
from requests.utils import parse_dict_header

from . import __version__
from . import exceptions
from .utils import expand_path

logger = logging.getLogger(__name__)
USERAGENT = f"vdirsyncer/{__version__}"

# 'hack' to prevent aiohttp from loading the netrc config,
# but still allow it to read PROXY_* env vars.
# Otherwise, if our host is defined in the netrc config,
# aiohttp will overwrite our Authorization header.
# https://github.com/pimutils/vdirsyncer/issues/1138
os.environ["NETRC"] = "NUL" if platform.system() == "Windows" else "/dev/null"


class AuthMethod(ABC):
    def __init__(self, username, password):
        self.username = username
        self.password = password

    @abstractmethod
    def handle_401(self, response):
        raise NotImplementedError

    @abstractmethod
    def get_auth_header(self, method, url):
        raise NotImplementedError

    def __eq__(self, other):
        if not isinstance(other, AuthMethod):
            return False
        return (
            self.__class__ == other.__class__
            and self.username == other.username
            and self.password == other.password
        )


class BasicAuthMethod(AuthMethod):
    def handle_401(self, _response):
        pass

    def get_auth_header(self, _method, _url):
        auth_str = f"{self.username}:{self.password}"
        return "Basic " + b64encode(auth_str.encode("utf-8")).decode("utf-8")


class DigestAuthMethod(AuthMethod):
    # make class var to 'cache' the state, which is more efficient because otherwise
    # each request would first require another 'initialization' request.
    _auth_helpers: dict[tuple[str, str], requests.auth.HTTPDigestAuth] = {}

    def __init__(self, username: str, password: str):
        super().__init__(username, password)

        self._auth_helper = self._auth_helpers.get(
            (username, password), requests.auth.HTTPDigestAuth(username, password)
        )
        self._auth_helpers[(username, password)] = self._auth_helper

    @property
    def auth_helper_vars(self):
        return self._auth_helper._thread_local

    def handle_401(self, response):
        s_auth = response.headers.get("www-authenticate", "")

        if "digest" in s_auth.lower():
            # Original source:
            # https://github.com/psf/requests/blob/f12ccbef6d6b95564da8d22e280d28c39d53f0e9/src/requests/auth.py#L262-L263
            pat = re.compile(r"digest ", flags=re.IGNORECASE)
            self.auth_helper_vars.chal = parse_dict_header(pat.sub("", s_auth, count=1))

    def get_auth_header(self, method, url):
        self._auth_helper.init_per_thread_state()

        if not self.auth_helper_vars.chal:
            # Need to do init request first
            return ""

        return self._auth_helper.build_digest_header(method, url)


def prepare_auth(auth, username, password):
    if username and password:
        if auth == "basic" or auth is None:
            return BasicAuthMethod(username, password)
        elif auth == "digest":
            return DigestAuthMethod(username, password)
        elif auth == "guess":
            raise exceptions.UserError(
                "'Guess' authentication is not supported in this version of vdirsyncer. \n"
                "Please explicitly specify either 'basic' or 'digest' auth instead. \n"
                "See the following issue for more information: "
                "https://github.com/pimutils/vdirsyncer/issues/1015"
            )
        else:
            raise exceptions.UserError(f"Unknown authentication method: {auth}")
    elif auth:
        raise exceptions.UserError(
            f"You need to specify username and password for {auth} authentication."
        )

    return None


def prepare_verify(verify, verify_fingerprint):
    if isinstance(verify, str):
        return create_default_context(cafile=expand_path(verify))
    elif verify is not None:
        raise exceptions.UserError(
            f"Invalid value for verify ({verify}), must be a path to a PEM-file."
        )

    if verify_fingerprint is not None:
        if not isinstance(verify_fingerprint, str):
            raise exceptions.UserError(
                "Invalid value for verify_fingerprint "
                f"({verify_fingerprint}), must be a string."
            )

        return aiohttp.Fingerprint(bytes.fromhex(verify_fingerprint.replace(":", "")))

    return None


def prepare_client_cert(cert):
    if isinstance(cert, (str, bytes)):
        cert = expand_path(cert)
    elif isinstance(cert, list):
        cert = tuple(map(prepare_client_cert, cert))
    return cert


async def request(
    method,
    url,
    session,
    auth=None,
    latin1_fallback=True,
    **kwargs,
):
    """Wrapper method for requests, to ease logging and mocking as well as to
    support auth methods currently unsupported by aiohttp.

    Parameters should be the same as for ``aiohttp.request``, except:

    :param session: A requests session object to use.
    :param auth: The HTTP ``AuthMethod`` to use for authentication.
    :param verify_fingerprint: Optional. SHA256 of the expected server certificate.
    :param latin1_fallback: RFC-2616 specifies the default Content-Type of
        text/* to be latin1, which is not always correct, but exactly what
        requests is doing. Setting this parameter to False will use charset
        autodetection (usually ending up with utf8) instead of plainly falling
        back to this silly default. See
        https://github.com/kennethreitz/requests/issues/2042
    """

    # TODO: Support for client-side certifications.

    session.hooks = {"response": _fix_redirects}

    # TODO: rewrite using
    # https://docs.aiohttp.org/en/stable/client_advanced.html#client-tracing
    logger.debug("=" * 20)
    logger.debug(f"{method} {url}")
    logger.debug(kwargs.get("headers", {}))
    logger.debug(kwargs.get("data", None))
    logger.debug("Sending request...")

    assert isinstance(kwargs.get("data", b""), bytes)

    cert = kwargs.pop("cert", None)
    if cert is not None:
        ssl_context = kwargs.pop("ssl", create_default_context())
        ssl_context.load_cert_chain(*cert)
        kwargs["ssl"] = ssl_context

    headers = kwargs.pop("headers", {})
    num_401 = 0
    while num_401 < 2:
        if auth:
            headers["Authorization"] = auth.get_auth_header(method, url)
        response = await session.request(method, url, headers=headers, **kwargs)

        if response.ok or not auth:
            # we don't need to do the 401-loop if we don't do auth in the first place
            break

        if response.status == 401:
            num_401 += 1
            auth.handle_401(response)
        else:
            # some other error, will be handled later on
            break

    # See https://github.com/kennethreitz/requests/issues/2042
    content_type = response.headers.get("Content-Type", "")
    if (
        not latin1_fallback
        and "charset" not in content_type
        and content_type.startswith("text/")
    ):
        logger.debug("Removing latin1 fallback")
        response.encoding = None

    logger.debug(response.status)
    logger.debug(response.headers)
    logger.debug(response.content)

    if logger.getEffectiveLevel() <= logging.DEBUG and response.status >= 400:
        # https://github.com/pimutils/vdirsyncer/issues/1186
        logger.debug(await response.text())

    if response.status == 412:
        raise exceptions.PreconditionFailed(response.reason)
    if response.status in (404, 410):
        raise exceptions.NotFoundError(response.reason)

    response.raise_for_status()
    return response


def _fix_redirects(r, *args, **kwargs):
    """
    Requests discards of the body content when it is following a redirect that
    is not a 307 or 308. We never want that to happen.

    See:
    https://github.com/kennethreitz/requests/issues/3915
    https://github.com/pimutils/vdirsyncer/pull/585
    https://github.com/pimutils/vdirsyncer/issues/586

    FIXME: This solution isn't very nice. A new hook in requests would be
    better.
    """
    if r.is_redirect:
        logger.debug("Rewriting status code from %s to 307", r.status_code)
        r.status_code = 307



from __future__ import annotations

import json
import os
import string
from configparser import RawConfigParser
from itertools import chain
from typing import IO
from typing import Any
from typing import Generator
from typing import Tuple

from .. import PROJECT_HOME
from .. import exceptions
from ..utils import cached_property
from ..utils import expand_path
from ..vobject import Item
from .fetchparams import expand_fetch_params
from .utils import storage_class_from_config

GENERAL_ALL = frozenset(["status_path"])
GENERAL_REQUIRED = frozenset(["status_path"])
SECTION_NAME_CHARS = frozenset(chain(string.ascii_letters, string.digits, "_"))


def validate_section_name(name, section_type):
    invalid = set(name) - SECTION_NAME_CHARS
    if invalid:
        chars_display = "".join(sorted(SECTION_NAME_CHARS))
        raise exceptions.UserError(
            f'The {section_type}-section "{name}" contains invalid characters. Only '
            "the following characters are allowed for storage and "
            f"pair names:\n{chars_display}"
        )


def _validate_general_section(general_config: dict[str, str]):
    invalid = set(general_config) - GENERAL_ALL
    missing = GENERAL_REQUIRED - set(general_config)
    problems: list[str] = []

    if invalid:
        problems.append(
            "general section doesn't take the parameters: {}".format(", ".join(invalid))
        )

    if missing:
        problems.append(
            "general section is missing the parameters: {}".format(", ".join(missing))
        )

    if problems:
        raise exceptions.UserError(
            "Invalid general section. Copy the example "
            f"config from the repository and edit it: {PROJECT_HOME}",
            problems=problems,
        )


def _validate_collections_param(collections):
    if collections is None:
        return

    if not isinstance(collections, list):
        raise ValueError("`collections` parameter must be a list or `null`.")

    collection_names = set()

    for i, collection in enumerate(collections):
        try:
            if isinstance(collection, (str, bytes)):
                collection_name = collection
            elif isinstance(collection, list):
                e = ValueError(
                    "Expected list of format "
                    '["config_name", "storage_a_name", "storage_b_name"]'
                )
                if len(collection) != 3:
                    raise e

                if not isinstance(collection[0], (str, bytes)):
                    raise e

                for x in collection[1:]:
                    if x is not None and not isinstance(x, (str, bytes)):
                        raise e

                collection_name = collection[0]
            else:
                raise ValueError("Expected string or list of three strings.")

            if collection_name in collection_names:
                raise ValueError("Duplicate value.")
            collection_names.add(collection_name)
        except ValueError as e:
            raise ValueError(f"`collections` parameter, position {i}: {str(e)}")


class _ConfigReader:
    def __init__(self, f: IO[Any]):
        self._file: IO[Any] = f
        self._parser = c = RawConfigParser()
        c.read_file(f)
        self._seen_names: set = set()

        self._general: dict[str, str] = {}
        self._pairs: dict[str, dict[str, str]] = {}
        self._storages: dict[str, dict[str, str]] = {}

    def _parse_section(
        self, section_type: str, name: str, options: dict[str, Any]
    ) -> None:
        validate_section_name(name, section_type)
        if name in self._seen_names:
            raise ValueError(f'Name "{name}" already used.')
        self._seen_names.add(name)

        if section_type == "general":
            if self._general:
                raise ValueError("More than one general section.")
            self._general = options
        elif section_type == "storage":
            self._storages[name] = options
        elif section_type == "pair":
            self._pairs[name] = options
        else:
            raise ValueError("Unknown section type.")

    def parse(
        self,
    ) -> tuple[dict[str, str], dict[str, dict[str, str]], dict[str, dict[str, str]]]:
        for section in self._parser.sections():
            if " " in section:
                section_type, name = section.split(" ", 1)
            else:
                section_type = name = section

            try:
                self._parse_section(
                    section_type,
                    name,
                    dict(_parse_options(self._parser.items(section), section=section)),
                )
            except ValueError as e:
                raise exceptions.UserError(f'Section "{section}": {str(e)}')

        _validate_general_section(self._general)
        if getattr(self._file, "name", None):
            self._general["status_path"] = os.path.join(
                os.path.dirname(self._file.name),
                expand_path(self._general["status_path"]),
            )

        return self._general, self._pairs, self._storages


def _parse_options(
    items: list[tuple[str, str]], section: str | None = None
) -> Generator[tuple[str, dict[str, str]], None, None]:
    for key, value in items:
        try:
            yield key, json.loads(value)
        except ValueError as e:
            raise ValueError(f'Section "{section}", option "{key}": {e}')


class Config:
    def __init__(
        self,
        general: dict[str, str],
        pairs: dict[str, dict[str, str]],
        storages: dict[str, dict[str, str]],
    ) -> None:
        self.general = general
        self.storages = storages
        for name, options in storages.items():
            options["instance_name"] = name

        self.pairs: dict[str, PairConfig] = {}
        for name, options in pairs.items():
            try:
                self.pairs[name] = PairConfig(self, name, options)
            except ValueError as e:
                raise exceptions.UserError(f"Pair {name}: {e}")

    @classmethod
    def from_fileobject(cls, f: IO[Any]):
        reader = _ConfigReader(f)
        return cls(*reader.parse())

    @classmethod
    def from_filename_or_environment(cls, fname: str | None = None):
        if fname is None:
            fname = os.environ.get("VDIRSYNCER_CONFIG", None)
        if fname is None:
            fname = expand_path("~/.vdirsyncer/config")
            if not os.path.exists(fname):
                xdg_config_dir = os.environ.get(
                    "XDG_CONFIG_HOME", expand_path("~/.config/")
                )
                fname = os.path.join(xdg_config_dir, "vdirsyncer/config")

        try:
            with open(fname) as f:
                return cls.from_fileobject(f)
        except Exception as e:
            raise exceptions.UserError(f"Error during reading config {fname}: {e}")

    def get_storage_args(self, storage_name: str):
        try:
            args = self.storages[storage_name]
        except KeyError:
            raise exceptions.UserError(
                f"Storage {storage_name!r} not found. "
                f"These are the configured storages: {list(self.storages)}"
            )
        else:
            return expand_fetch_params(args)

    def get_pair(self, pair_name: str) -> PairConfig:
        try:
            return self.pairs[pair_name]
        except KeyError as e:
            raise exceptions.PairNotFound(e, pair_name=pair_name)


class PairConfig:
    def __init__(self, full_config: Config, name: str, options: dict[str, str]):
        self._config: Config = full_config
        self.name: str = name
        self.name_a: str = options.pop("a")
        self.name_b: str = options.pop("b")

        self._partial_sync: str | None = options.pop("partial_sync", None)
        self.metadata: str | Tuple[()] = options.pop("metadata", ())

        self.conflict_resolution = self._process_conflict_resolution_param(
            options.pop("conflict_resolution", None)
        )

        try:
            self.collections = options.pop("collections")
        except KeyError:
            raise ValueError(
                "collections parameter missing.\n\n"
                "As of 0.9.0 this parameter has no default anymore. "
                "Set `collections = null` explicitly in your pair config."
            )
        else:
            _validate_collections_param(self.collections)

        if options:
            raise ValueError("Unknown options: {}".format(", ".join(options)))

    def _process_conflict_resolution_param(
        self, conflict_resolution: str | list[str] | None
    ):
        if conflict_resolution in (None, "a wins", "b wins"):
            return conflict_resolution
        elif (
            isinstance(conflict_resolution, list)
            and len(conflict_resolution) > 1
            and conflict_resolution[0] == "command"
        ):

            def resolve(a, b):
                a_name = self.config_a["instance_name"]
                b_name = self.config_b["instance_name"]
                command = conflict_resolution[1:]

                return _resolve_conflict_via_command(a, b, command, a_name, b_name)

            return resolve
        else:
            raise ValueError("Invalid value for `conflict_resolution`.")

    # The following parameters are lazily evaluated because evaluating
    # self.config_a would expand all `x.fetch` parameters. This is costly and
    # unnecessary if the pair is not actually synced.

    @cached_property
    def config_a(self):
        return self._config.get_storage_args(self.name_a)

    @cached_property
    def config_b(self):
        return self._config.get_storage_args(self.name_b)

    @cached_property
    def partial_sync(self):
        partial_sync = self._partial_sync
        # We need to use UserError here because ValueError is not
        # caught at the time this is expanded.

        if partial_sync is not None:
            cls_a, _ = storage_class_from_config(self.config_a)
            cls_b, _ = storage_class_from_config(self.config_b)

            if (
                not cls_a.read_only
                and not self.config_a.get("read_only", False)
                and not cls_b.read_only
                and not self.config_b.get("read_only", False)
            ):
                raise exceptions.UserError(
                    "`partial_sync` is only effective if one storage is "
                    "read-only. Use `read_only = true` in exactly one storage "
                    "section."
                )

        if partial_sync is None:
            partial_sync = "revert"

        if partial_sync not in ("ignore", "revert", "error"):
            raise exceptions.UserError("Invalid value for `partial_sync`.")

        return partial_sync


class CollectionConfig:
    def __init__(self, pair, name: str, config_a, config_b):
        self.pair = pair
        self._config = pair._config
        self.name: str = name
        self.config_a = config_a
        self.config_b = config_b


#: Public API. Khal's config wizard depends on this function.
load_config = Config.from_filename_or_environment


def _resolve_conflict_via_command(
    a, b, command, a_name, b_name, _check_call=None
) -> Item:
    import shutil
    import tempfile

    if _check_call is None:
        from subprocess import check_call as _check_call

    from ..vobject import Item

    dir = tempfile.mkdtemp(prefix="vdirsyncer-conflict.")
    try:
        a_tmp = os.path.join(dir, a_name)
        b_tmp = os.path.join(dir, b_name)

        with open(a_tmp, "w") as f:
            f.write(a.raw)
        with open(b_tmp, "w") as f:
            f.write(b.raw)

        command[0] = expand_path(command[0])
        _check_call(command + [a_tmp, b_tmp])

        with open(a_tmp) as f:
            new_a = f.read()
        with open(b_tmp) as f:
            new_b = f.read()

        if new_a != new_b:
            raise exceptions.UserError("The two files are not completely equal.")
        return Item(new_a)
    finally:
        shutil.rmtree(dir)



from __future__ import annotations

from .. import exceptions


class SyncError(exceptions.Error):
    """Errors related to synchronization."""


class SyncConflict(SyncError):
    """
    Two items changed since the last sync, they now have different contents and
    no conflict resolution method was given.

    :param ident: The ident of the item.
    :param href_a: The item's href on side A.
    :param href_b: The item's href on side B.
    """

    ident = None
    href_a = None
    href_b = None


class IdentConflict(SyncError):
    """
    Multiple items on the same storage have the same UID.

    :param storage: The affected storage.
    :param hrefs: List of affected hrefs on `storage`.
    """

    storage = None
    _hrefs = None

    @property
    def hrefs(self):
        return self._hrefs

    @hrefs.setter
    def hrefs(self, val):
        new_val = set(val)
        assert len(new_val) > 1, val
        self._hrefs = new_val


class StorageEmpty(SyncError):
    """
    One storage unexpectedly got completely empty between two synchronizations.
    The first argument is the empty storage.

    :param empty_storage: The empty
        :py:class:`vdirsyncer.storage.base.Storage`.
    """

    empty_storage = None


class BothReadOnly(SyncError):
    """
    Both storages are marked as read-only. Synchronization is therefore not
    possible.
    """


class PartialSync(SyncError):
    """
    Attempted change on read-only storage.
    """

    storage = None


class IdentAlreadyExists(SyncError):
    """Like IdentConflict, but for internal state. If this bubbles up, we don't
    have a data race, but a bug."""

    old_href = None
    new_href = None

    def to_ident_conflict(self, storage):
        return IdentConflict(storage=storage, hrefs=[self.old_href, self.new_href])



from __future__ import annotations

import random

from .. import exceptions
from .base import Storage
from .base import normalize_meta_value


def _random_string():
    return f"{random.random():.9f}"


class MemoryStorage(Storage):
    storage_name = "memory"

    """
    Saves data in RAM, only useful for testing.
    """

    def __init__(self, fileext="", **kwargs):
        if kwargs.get("collection") is not None:
            raise exceptions.UserError("MemoryStorage does not support collections.")
        self.items = {}  # href => (etag, item)
        self.metadata = {}
        self.fileext = fileext
        super().__init__(**kwargs)

    def _get_href(self, item):
        return item.ident + self.fileext

    async def list(self):
        for href, (etag, _item) in self.items.items():
            yield href, etag

    async def get(self, href):
        etag, item = self.items[href]
        return item, etag

    async def has(self, href):
        return href in self.items

    async def upload(self, item):
        href = self._get_href(item)
        if href in self.items:
            raise exceptions.AlreadyExistingError(existing_href=href)
        etag = _random_string()
        self.items[href] = (etag, item)
        return href, etag

    async def update(self, href, item, etag):
        if href not in self.items:
            raise exceptions.NotFoundError(href)
        actual_etag, _ = self.items[href]
        if etag != actual_etag:
            raise exceptions.WrongEtagError(etag, actual_etag)

        new_etag = _random_string()
        self.items[href] = (new_etag, item)
        return new_etag

    async def delete(self, href, etag):
        if not await self.has(href):
            raise exceptions.NotFoundError(href)
        if etag != self.items[href][0]:
            raise exceptions.WrongEtagError(etag)
        del self.items[href]

    async def get_meta(self, key):
        return normalize_meta_value(self.metadata.get(key))

    async def set_meta(self, key, value):
        if value is None:
            self.metadata.pop(key, None)
        else:
            self.metadata[key] = normalize_meta_value(value)



from __future__ import annotations

import logging

import click

from .. import exceptions
from ..utils import expand_path
from ..utils import synchronized
from . import AppContext

SUFFIX = ".fetch"

logger = logging.getLogger(__name__)


def expand_fetch_params(config):
    config = dict(config)
    for key in list(config):
        if not key.endswith(SUFFIX):
            continue

        newkey = key[: -len(SUFFIX)]
        if newkey in config:
            raise ValueError(f"Can't set {key} and {newkey}.")
        config[newkey] = _fetch_value(config[key], key)
        del config[key]

    return config


@synchronized()
def _fetch_value(opts, key):
    if not isinstance(opts, list):
        raise ValueError(f"Invalid value for {key}: Expected a list, found {opts!r}.")
    if not opts:
        raise ValueError("Expected list of length > 0.")

    try:
        ctx = click.get_current_context().find_object(AppContext)
        if ctx is None:
            raise RuntimeError
        password_cache = ctx.fetched_params
    except RuntimeError:
        password_cache = {}

    cache_key = tuple(opts)
    if cache_key in password_cache:
        rv = password_cache[cache_key]
        logger.debug(f"Found cached value for {opts!r}.")
        if isinstance(rv, BaseException):
            raise rv
        return rv

    strategy = opts[0]
    try:
        strategy_fn = STRATEGIES[strategy]
    except KeyError:
        raise exceptions.UserError(f"Unknown strategy: {strategy}")

    logger.debug(f"Fetching value for {key} with {strategy} strategy.")
    try:
        rv = strategy_fn(*opts[1:])
    except (click.Abort, KeyboardInterrupt) as e:
        password_cache[cache_key] = e
        raise
    else:
        if not rv:
            raise exceptions.UserError(
                f"Empty value for {key}, this most likely indicates an error."
            )
        password_cache[cache_key] = rv
        return rv


def _strategy_command(*command: str, shell: bool = False):
    """Execute a user-specified command and return its output."""
    import subprocess

    # Normalize path of every path member.
    # If there is no path specified then nothing will happen.
    # Makes this a list to avoid it being exhausted on the first iteration.
    expanded_command = list(map(expand_path, command))

    try:
        stdout = subprocess.check_output(expanded_command, text=True, shell=shell)
        return stdout.strip("\n")
    except OSError as e:
        cmd = " ".join(expanded_command)
        raise exceptions.UserError(f"Failed to execute command: {cmd}\n{str(e)}")


def _strategy_shell(*command: str):
    """Execute a user-specified command string in a shell and return its output."""
    return _strategy_command(*command, shell=True)


def _strategy_prompt(text):
    return click.prompt(text, hide_input=True)


STRATEGIES = {
    "command": _strategy_command,
    "shell": _strategy_shell,
    "prompt": _strategy_prompt,
}



from __future__ import annotations

import json
import logging
import os
import urllib.parse as urlparse
import wsgiref.simple_server
import wsgiref.util
from pathlib import Path
from threading import Thread

import aiohttp
import click

from .. import exceptions
from ..utils import atomic_write
from ..utils import checkdir
from ..utils import expand_path
from ..utils import open_graphical_browser
from . import base
from . import dav
from .google_helpers import _RedirectWSGIApp
from .google_helpers import _WSGIRequestHandler

logger = logging.getLogger(__name__)


TOKEN_URL = "https://accounts.google.com/o/oauth2/v2/auth"
REFRESH_URL = "https://www.googleapis.com/oauth2/v4/token"

try:
    from aiohttp_oauthlib import OAuth2Session

    have_oauth2 = True
except ImportError:
    have_oauth2 = False


class GoogleSession(dav.DAVSession):
    def __init__(
        self,
        token_file,
        client_id,
        client_secret,
        url=None,
        *,
        connector: aiohttp.BaseConnector,
    ):
        if not have_oauth2:
            raise exceptions.UserError("aiohttp-oauthlib not installed")

        # Required for discovering collections
        if url is not None:
            self.url = url

        self.useragent = client_id
        self._settings = {}
        self.connector = connector

        self._token_file = Path(expand_path(token_file))
        self._client_id = client_id
        self._client_secret = client_secret
        self._token = None
        self._redirect_uri = None

    async def request(self, method, path, **kwargs):
        if not self._token:
            await self._init_token()

        return await super().request(method, path, **kwargs)

    async def _save_token(self, token):
        """Helper function called by OAuth2Session when a token is updated."""
        checkdir(expand_path(os.path.dirname(self._token_file)), create=True)
        with atomic_write(self._token_file, mode="w", overwrite=True) as f:
            json.dump(token, f)

    @property
    def _session(self):
        """Return a new OAuth session for requests.

        Accesses the self.redirect_uri field (str): the URI to redirect
        authentication to. Should be a loopback address for a local server that
        follows the process detailed in
        https://developers.google.com/identity/protocols/oauth2/native-app.
        """

        return OAuth2Session(
            client_id=self._client_id,
            token=self._token,
            redirect_uri=self._redirect_uri,
            scope=self.scope,
            auto_refresh_url=REFRESH_URL,
            auto_refresh_kwargs={
                "client_id": self._client_id,
                "client_secret": self._client_secret,
            },
            token_updater=self._save_token,
            connector=self.connector,
            connector_owner=False,
            trust_env=True,
        )

    async def _init_token(self):
        try:
            with self._token_file.open() as f:
                self._token = json.load(f)
        except FileNotFoundError:
            pass
        except ValueError as e:
            raise exceptions.UserError(
                f"Failed to load token file {self._token_file}, try deleting it. "
                f"Original error: {e}"
            )

        if not self._token:
            # Some times a task stops at this `async`, and another continues the flow.
            # At this point, the user has already completed the flow, but is prompeted
            # for a second one.
            wsgi_app = _RedirectWSGIApp("Successfully obtained token.")
            wsgiref.simple_server.WSGIServer.allow_reuse_address = False
            host = "127.0.0.1"
            local_server = wsgiref.simple_server.make_server(
                host, 0, wsgi_app, handler_class=_WSGIRequestHandler
            )
            thread = Thread(target=local_server.handle_request)
            thread.start()
            self._redirect_uri = f"http://{host}:{local_server.server_port}"
            async with self._session as session:
                # Fail fast if the address is occupied

                authorization_url, state = session.authorization_url(
                    TOKEN_URL,
                    # access_type and approval_prompt are Google specific
                    # extra parameters.
                    access_type="offline",
                    approval_prompt="force",
                )
                click.echo(f"Opening {authorization_url} ...")
                try:
                    open_graphical_browser(authorization_url)
                except Exception as e:
                    logger.warning(str(e))

                click.echo("Follow the instructions on the page.")
                thread.join()
                logger.debug("server handled request!")

                # Note: using https here because oauthlib is very picky that
                # OAuth 2.0 should only occur over https.
                authorization_response = wsgi_app.last_request_uri.replace(
                    "http", "https", 1
                )
                logger.debug(f"authorization_response: {authorization_response}")
                self._token = await session.fetch_token(
                    REFRESH_URL,
                    authorization_response=authorization_response,
                    # Google specific extra param used for client authentication:
                    client_secret=self._client_secret,
                )
                logger.debug(f"token: {self._token}")
                local_server.server_close()

            # FIXME: Ugly
            await self._save_token(self._token)


class GoogleCalendarStorage(dav.CalDAVStorage):
    class session_class(GoogleSession):
        url = "https://apidata.googleusercontent.com/caldav/v2/"
        scope = ["https://www.googleapis.com/auth/calendar"]

    class discovery_class(dav.CalDiscover):
        @staticmethod
        def _get_collection_from_url(url):
            # Google CalDAV has collection URLs like:
            # /user/foouser/calendars/foocalendar/events/
            parts = url.rstrip("/").split("/")
            parts.pop()
            collection = parts.pop()
            return urlparse.unquote(collection)

    storage_name = "google_calendar"

    def __init__(
        self,
        token_file,
        client_id,
        client_secret,
        start_date=None,
        end_date=None,
        item_types=(),
        **kwargs,
    ):
        if not kwargs.get("collection"):
            raise exceptions.CollectionRequired

        super().__init__(
            token_file=token_file,
            client_id=client_id,
            client_secret=client_secret,
            start_date=start_date,
            end_date=end_date,
            item_types=item_types,
            **kwargs,
        )

    # This is ugly: We define/override the entire signature computed for the
    # docs here because the current way we autogenerate those docs are too
    # simple for our advanced argspec juggling in `vdirsyncer.storage.dav`.
    __init__._traverse_superclass = base.Storage  # type: ignore


class GoogleContactsStorage(dav.CardDAVStorage):
    class session_class(GoogleSession):
        # Google CardDAV is completely bonkers. Collection discovery doesn't
        # work properly, well-known URI takes us directly to single collection
        # from where we can't discover principal or homeset URIs (the PROPFINDs
        # 404).
        #
        # So we configure the well-known URI here again, such that discovery
        # tries collection enumeration on it directly. That appears to work.
        url = "https://www.googleapis.com/.well-known/carddav"
        scope = ["https://www.googleapis.com/auth/carddav"]

    class discovery_class(dav.CardDiscover):
        # Google CardDAV doesn't return any resourcetype prop.
        _resourcetype = None

    storage_name = "google_contacts"

    def __init__(self, token_file, client_id, client_secret, **kwargs):
        if not kwargs.get("collection"):
            raise exceptions.CollectionRequired

        super().__init__(
            token_file=token_file,
            client_id=client_id,
            client_secret=client_secret,
            **kwargs,
        )

    # This is ugly: We define/override the entire signature computed for the
    # docs here because the current way we autogenerate those docs are too
    # simple for our advanced argspec juggling in `vdirsyncer.storage.dav`.
    __init__._traverse_superclass = base.Storage  # type: ignore



# SPDX-License-Identifier: Apache-2.0
#
# Based on:
# https://github.com/googleapis/google-auth-library-python-oauthlib/blob/1fb16be1bad9050ee29293541be44e41e82defd7/google_auth_oauthlib/flow.py#L513
from __future__ import annotations

import logging
import wsgiref.simple_server
import wsgiref.util
from typing import Any
from typing import Callable
from typing import Iterable

logger = logging.getLogger(__name__)


class _WSGIRequestHandler(wsgiref.simple_server.WSGIRequestHandler):
    """Custom WSGIRequestHandler."""

    def log_message(self, format, *args):
        # (format is the argument name defined in the superclass.)
        logger.info(format, *args)


class _RedirectWSGIApp:
    """WSGI app to handle the authorization redirect.

    Stores the request URI and displays the given success message.
    """

    last_request_uri: str | None

    def __init__(self, success_message: str):
        """
        :param success_message: The message to display in the web browser the
            authorization flow is complete.
        """
        self.last_request_uri = None
        self._success_message = success_message

    def __call__(
        self,
        environ: dict[str, Any],
        start_response: Callable[[str, list], None],
    ) -> Iterable[bytes]:
        """WSGI Callable.

        :param environ: The WSGI environment.
        :param start_response: The WSGI start_response callable.
        :returns: The response body.
        """
        start_response("200 OK", [("Content-type", "text/plain; charset=utf-8")])
        self.last_request_uri = wsgiref.util.request_uri(environ)
        return [self._success_message.encode("utf-8")]



"""
There are storage classes which control the access to one vdir-collection and
offer basic CRUD-ish methods for modifying those collections. The exact
interface is described in `vdirsyncer.storage.base`, the `Storage` class should
be a superclass of all storage classes.
"""



from __future__ import annotations

import logging
from os.path import basename

import aiostream

from .utils import generate_href
from .utils import href_safe

logger = logging.getLogger(__name__)


class IrreparableItem(Exception):
    pass


async def repair_storage(storage, repair_unsafe_uid):
    seen_uids = set()
    all_hrefs = await aiostream.stream.list(storage.list())
    for i, (href, _) in enumerate(all_hrefs):
        item, etag = await storage.get(href)
        logger.info(f"[{i}/{len(all_hrefs)}] Processing {href}")

        try:
            new_item = repair_item(href, item, seen_uids, repair_unsafe_uid)
        except IrreparableItem:
            logger.error(
                f"Item {href!r} is malformed beyond repair. "
                "The PRODID property may indicate which software "
                "created this item."
            )
            logger.error(f"Item content: {item.raw!r}")
            continue

        seen_uids.add(new_item.uid)
        if new_item.raw != item.raw:
            if new_item.uid != item.uid:
                await storage.upload(new_item)
                await storage.delete(href, etag)
            else:
                await storage.update(href, new_item, etag)


def repair_item(href, item, seen_uids, repair_unsafe_uid):
    if item.parsed is None:
        raise IrreparableItem

    new_item = item

    if not item.uid:
        logger.warning("No UID, assigning random UID.")
        new_item = item.with_uid(generate_href())
    elif item.uid in seen_uids:
        logger.warning("Duplicate UID, assigning random UID.")
        new_item = item.with_uid(generate_href())
    elif not href_safe(item.uid) or not href_safe(basename(href)):
        if not repair_unsafe_uid:
            logger.warning("UID may cause problems, add --repair-unsafe-uid to repair.")
        else:
            logger.warning("UID or href is unsafe, assigning random UID.")
            new_item = item.with_uid(generate_href())

    if not new_item.uid:
        raise IrreparableItem

    return new_item



from __future__ import annotations

import asyncio
import functools
import json
import logging
import sys

import aiohttp
import click
import click_log

from .. import BUGTRACKER_HOME
from .. import __version__

cli_logger = logging.getLogger(__name__)
click_log.basic_config("vdirsyncer")

# add short option for the help option
click_context_settings = {"help_option_names": ["-h", "--help"]}


class AppContext:
    def __init__(self):
        self.config = None
        self.fetched_params = {}
        self.logger = None


pass_context = click.make_pass_decorator(AppContext, ensure=True)


def catch_errors(f):
    @functools.wraps(f)
    def inner(*a, **kw):
        try:
            f(*a, **kw)
        except BaseException:
            from .utils import handle_cli_error

            handle_cli_error()
            sys.exit(1)

    return inner


@click.group(context_settings=click_context_settings)
@click_log.simple_verbosity_option("vdirsyncer")
@click.version_option(version=__version__)
@click.option("--config", "-c", metavar="FILE", help="Config file to use.")
@pass_context
@catch_errors
def app(ctx, config: str):
    """
    Synchronize calendars and contacts
    """

    if sys.platform == "win32":
        cli_logger.warning(
            "Vdirsyncer currently does not support Windows. "
            "You will likely encounter bugs. "
            f"See {BUGTRACKER_HOME}/535 for more information."
        )

    if not ctx.config:
        from .config import load_config

        ctx.config = load_config(config)


def collections_arg_callback(ctx, param, value):
    """
    Expand the various CLI shortforms ("pair, pair/collection") to an iterable
    of (pair, collections).
    """
    # XXX: Ugly! pass_context should work everywhere.
    config = ctx.find_object(AppContext).config
    rv = {}
    for pair_and_collection in value or config.pairs:
        pair, collection = pair_and_collection, None
        if "/" in pair:
            pair, collection = pair.split("/")

        collections = rv.setdefault(pair, set())
        if collection:
            collections.add(collection)

    return rv.items()


collections_arg = click.argument(
    "collections", nargs=-1, callback=collections_arg_callback
)


@app.command()
@collections_arg
@click.option(
    "--force-delete/--no-force-delete",
    help=(
        "Do/Don't abort synchronization when all items are about "
        "to be deleted from both sides."
    ),
)
@pass_context
@catch_errors
def sync(ctx, collections, force_delete):
    """
    Synchronize the given collections or pairs. If no arguments are given, all
    will be synchronized.

    This command will not synchronize metadata, use `vdirsyncer metasync` for
    that.

    \b
    \b\bExamples:
    # Sync everything configured
    vdirsyncer sync

    \b
    # Sync the pairs "bob" and "frank"
    vdirsyncer sync bob frank

    \b
    # Sync only "first_collection" from the pair "bob"
    vdirsyncer sync bob/first_collection
    """
    from .tasks import prepare_pair
    from .tasks import sync_collection

    async def main(collection_names):
        async with aiohttp.TCPConnector(limit_per_host=16) as conn:
            tasks = []
            for pair_name, collections in collection_names:
                async for collection, config in prepare_pair(
                    pair_name=pair_name,
                    collections=collections,
                    config=ctx.config,
                    connector=conn,
                ):
                    tasks.append(
                        sync_collection(
                            collection=collection,
                            general=config,
                            force_delete=force_delete,
                            connector=conn,
                        )
                    )

            await asyncio.gather(*tasks)

    asyncio.run(main(collections))


@app.command()
@collections_arg
@pass_context
@catch_errors
def metasync(ctx, collections):
    """
    Synchronize metadata of the given collections or pairs.

    See the `sync` command for usage.
    """
    from .tasks import metasync_collection
    from .tasks import prepare_pair

    async def main(collection_names):
        async with aiohttp.TCPConnector(limit_per_host=16) as conn:
            for pair_name, collections in collection_names:
                collections = prepare_pair(
                    pair_name=pair_name,
                    collections=collections,
                    config=ctx.config,
                    connector=conn,
                )

                await asyncio.gather(
                    *[
                        metasync_collection(
                            collection=collection,
                            general=config,
                            connector=conn,
                        )
                        async for collection, config in collections
                    ]
                )

    asyncio.run(main(collections))


@app.command()
@click.argument("pairs", nargs=-1)
@click.option(
    "--list/--no-list",
    default=True,
    help=(
        "Whether to list all collections from both sides during discovery, "
        "for debugging. This is slow and may crash for broken servers."
    ),
)
@pass_context
@catch_errors
def discover(ctx, pairs, list):
    """
    Refresh collection cache for the given pairs.
    """
    from .tasks import discover_collections

    config = ctx.config

    async def main():
        async with aiohttp.TCPConnector(limit_per_host=16) as conn:
            for pair_name in pairs or config.pairs:
                await discover_collections(
                    status_path=config.general["status_path"],
                    pair=config.get_pair(pair_name),
                    from_cache=False,
                    list_collections=list,
                    connector=conn,
                )

    asyncio.run(main())


@app.command()
@click.argument("collection")
@click.option(
    "--repair-unsafe-uid/--no-repair-unsafe-uid",
    default=False,
    help=(
        "Some characters in item UIDs and URLs may cause problems "
        "with buggy software. Adding this option will reassign "
        "new UIDs to those items. This is disabled by default, "
        "which is equivalent to `--no-repair-unsafe-uid`."
    ),
)
@pass_context
@catch_errors
def repair(ctx, collection, repair_unsafe_uid):
    """
    Repair a given collection.

    Runs a few checks on the collection and applies some fixes to individual
    items that may improve general stability, also with other CalDAV/CardDAV
    clients. In particular, if you encounter URL-encoding-related issues with
    other clients, this command with --repair-unsafe-uid might help.

    \b
    \b\bExamples:
    # Repair the `foo` collection of the `calendars_local` storage
    vdirsyncer repair calendars_local/foo
    """
    from .tasks import repair_collection

    cli_logger.warning("This operation will take a very long time.")
    cli_logger.warning(
        "It's recommended to make a backup and "
        "turn off other client's synchronization features."
    )
    click.confirm("Do you want to continue?", abort=True)

    async def main():
        async with aiohttp.TCPConnector(limit_per_host=16) as conn:
            await repair_collection(
                ctx.config,
                collection,
                repair_unsafe_uid=repair_unsafe_uid,
                connector=conn,
            )

    asyncio.run(main())


@app.command()
@pass_context
@catch_errors
def showconfig(ctx: AppContext):
    """Show the current configuration.

    This is mostly intended to be used by scripts or other integrations.
    If you need additional information in this dump, please reach out.
    """
    config = {"storages": list(ctx.config.storages.values())}
    click.echo(json.dumps(config, indent=2))



"""
Contains exception classes used by vdirsyncer. Not all exceptions are here,
only the most commonly used ones.
"""

from __future__ import annotations


class Error(Exception):
    """Baseclass for all errors."""

    def __init__(self, *args, **kwargs):
        for key, value in kwargs.items():
            if getattr(self, key, object()) is not None:  # pragma: no cover
                raise TypeError(f"Invalid argument: {key}")
            setattr(self, key, value)

        super().__init__(*args)


class UserError(Error, ValueError):
    """Wrapper exception to be used to signify the traceback should not be
    shown to the user."""

    problems = None

    def __str__(self):
        msg = Error.__str__(self)
        for problem in self.problems or ():
            msg += f"\n  - {problem}"

        return msg


class CollectionNotFound(Error):
    """Collection not found"""


class PairNotFound(Error):
    """Pair not found"""

    pair_name = None


class PreconditionFailed(Error):
    """
      - The item doesn't exist although it should
      - The item exists although it shouldn't
      - The etags don't match.

    Due to CalDAV we can't actually say which error it is.
    This error may indicate race conditions.
    """


class NotFoundError(PreconditionFailed):
    """Item not found"""


class AlreadyExistingError(PreconditionFailed):
    """Item already exists."""

    existing_href = None


class WrongEtagError(PreconditionFailed):
    """Wrong etag"""


class ReadOnlyError(Error):
    """Storage is read-only."""


class InvalidResponse(Error, ValueError):
    """The backend returned an invalid result."""


class UnsupportedMetadataError(Error, NotImplementedError):
    """The storage doesn't support this type of metadata."""


class CollectionRequired(Error):
    """`collection = null` is not allowed."""



from __future__ import annotations

import hashlib
from itertools import chain
from itertools import tee

from .utils import cached_property
from .utils import uniq

IGNORE_PROPS = (
    # PRODID is changed by radicale for some reason after upload
    "PRODID",
    # Sometimes METHOD:PUBLISH is added by WebCAL providers, for us it doesn't
    # make a difference
    "METHOD",
    # X-RADICALE-NAME is used by radicale, because hrefs don't really exist in
    # their filesystem backend
    "X-RADICALE-NAME",
    # Apparently this is set by Horde?
    # https://github.com/pimutils/vdirsyncer/issues/318
    "X-WR-CALNAME",
    # Those are from the VCARD specification and is supposed to change when the
    # item does -- however, we can determine that ourselves
    "REV",
    "LAST-MODIFIED",
    "CREATED",
    # Some iCalendar HTTP calendars generate the DTSTAMP at request time, so
    # this property always changes when the rest of the item didn't. Some do
    # the same with the UID.
    #
    # - Google's read-only calendar links
    # - http://www.feiertage-oesterreich.at/
    "DTSTAMP",
    "UID",
)


class Item:
    """Immutable wrapper class for VCALENDAR (VEVENT, VTODO) and
    VCARD"""

    def __init__(self, raw):
        assert isinstance(raw, str), type(raw)
        self._raw = raw

    def with_uid(self, new_uid):
        parsed = _Component.parse(self.raw)
        stack = [parsed]
        while stack:
            component = stack.pop()
            stack.extend(component.subcomponents)

            if component.name in ("VEVENT", "VTODO", "VJOURNAL", "VCARD"):
                del component["UID"]
                if new_uid:
                    component["UID"] = new_uid

        return Item("\r\n".join(parsed.dump_lines()))

    @cached_property
    def raw(self):
        """Raw content of the item, as unicode string.

        Vdirsyncer doesn't validate the content in any way.
        """
        return self._raw

    @cached_property
    def uid(self):
        """Global identifier of the item, across storages, doesn't change after
        a modification of the item."""
        # Don't actually parse component, but treat all lines as single
        # component, avoiding traversal through all subcomponents.
        x = _Component("TEMP", self.raw.splitlines(), [])
        try:
            return x["UID"].strip() or None
        except KeyError:
            return None

    @cached_property
    def hash(self):
        """Hash of self.raw, used for etags."""
        return hash_item(self.raw)

    @cached_property
    def ident(self):
        """Used for generating hrefs and matching up items during
        synchronization. This is either the UID or the hash of the item's
        content."""

        # We hash the item instead of directly using its raw content, because
        #
        # 1. The raw content might be really large, e.g. when it's a contact
        #    with a picture, which bloats the status file.
        #
        # 2. The status file would contain really sensitive information.
        return self.uid or self.hash

    @property
    def parsed(self):
        """Don't cache because the rv is mutable."""
        try:
            return _Component.parse(self.raw)
        except Exception:
            return None


def normalize_item(item, ignore_props=IGNORE_PROPS):
    """Create syntactically invalid mess that is equal for similar items."""
    if not isinstance(item, Item):
        item = Item(item)

    item = _strip_timezones(item)

    x = _Component("TEMP", item.raw.splitlines(), [])
    for prop in IGNORE_PROPS:
        del x[prop]

    x.props.sort()
    return "\r\n".join(filter(bool, (line.strip() for line in x.props)))


def _strip_timezones(item):
    parsed = item.parsed
    if not parsed or parsed.name != "VCALENDAR":
        return item

    parsed.subcomponents = [c for c in parsed.subcomponents if c.name != "VTIMEZONE"]

    return Item("\r\n".join(parsed.dump_lines()))


def hash_item(text):
    return hashlib.sha256(normalize_item(text).encode("utf-8")).hexdigest()


def split_collection(text):
    assert isinstance(text, str)
    inline = []
    items = {}  # uid => item
    ungrouped_items = []

    for main in _Component.parse(text, multiple=True):
        _split_collection_impl(main, main, inline, items, ungrouped_items)

    for item in chain(items.values(), ungrouped_items):
        item.subcomponents.extend(inline)
        yield "\r\n".join(item.dump_lines())


def _split_collection_impl(item, main, inline, items, ungrouped_items):
    if item.name == "VTIMEZONE":
        inline.append(item)
    elif item.name == "VCARD":
        ungrouped_items.append(item)
    elif item.name in ("VTODO", "VEVENT", "VJOURNAL"):
        uid = item.get("UID", "")
        wrapper = _Component(main.name, main.props[:], [])

        if uid.strip():
            wrapper = items.setdefault(uid, wrapper)
        else:
            ungrouped_items.append(wrapper)

        wrapper.subcomponents.append(item)
    elif item.name in ("VCALENDAR", "VADDRESSBOOK"):
        if item.name == "VCALENDAR":
            del item["METHOD"]
        for subitem in item.subcomponents:
            _split_collection_impl(subitem, item, inline, items, ungrouped_items)
    else:
        raise ValueError(f"Unknown component: {item.name}")


_default_join_wrappers = {
    "VCALENDAR": "VCALENDAR",
    "VEVENT": "VCALENDAR",
    "VTODO": "VCALENDAR",
    "VCARD": "VADDRESSBOOK",
}


def join_collection(items, wrappers=_default_join_wrappers):
    """
    :param wrappers: {
        item_type: wrapper_type
    }
    """

    items1, items2 = tee((_Component.parse(x) for x in items), 2)
    item_type, wrapper_type = _get_item_type(items1, wrappers)
    wrapper_props = []

    def _get_item_components(x):
        if x.name == wrapper_type:
            wrapper_props.extend(x.props)
            return x.subcomponents
        else:
            return [x]

    components = chain(*(_get_item_components(x) for x in items2))
    lines = chain(*uniq(tuple(x.dump_lines()) for x in components))

    if wrapper_type is not None:
        lines = chain(
            *(
                [f"BEGIN:{wrapper_type}"],
                # XXX: wrapper_props is a list of lines (with line-wrapping), so
                # filtering out duplicate lines will almost certainly break
                # multiline-values.  Since the only props we usually need to
                # support are PRODID and VERSION, I don't care.
                uniq(wrapper_props),
                lines,
                [f"END:{wrapper_type}"],
            )
        )
    return "".join(line + "\r\n" for line in lines)


def _get_item_type(components, wrappers):
    i = 0
    for component in components:
        i += 1
        try:
            item_type = component.name
            wrapper_type = wrappers[item_type]
        except KeyError:
            pass
        else:
            return item_type, wrapper_type

    if not i:
        return None, None
    else:
        raise ValueError("Not sure how to join components.")


class _Component:
    """
    Raw outline of the components.

    Vdirsyncer's operations on iCalendar and VCard objects are limited to
    retrieving the UID and splitting larger files into items. Consequently this
    parser is very lazy, with the downside that manipulation of item properties
    are extremely costly.

    Other features:

    - Preserve the original property order and wrapping.
    - Don't choke on irrelevant details like invalid datetime formats.

    Original version from https://github.com/collective/icalendar/, but apart
    from the similar API, very few parts have been reused.
    """

    def __init__(self, name, lines, subcomponents):
        """
        :param name: The component name.
        :param lines: The component's own properties, as list of lines
            (strings).
        :param subcomponents: List of components.
        """
        self.name = name
        self.props = lines
        self.subcomponents = subcomponents

    @classmethod
    def parse(cls, lines, multiple=False):
        if isinstance(lines, bytes):
            lines = lines.decode("utf-8")
        if isinstance(lines, str):
            lines = lines.splitlines()

        stack = []
        rv = []
        try:
            for _i, line in enumerate(lines):
                if line.startswith("BEGIN:"):
                    c_name = line[len("BEGIN:") :].strip().upper()
                    stack.append(cls(c_name, [], []))
                elif line.startswith("END:"):
                    component = stack.pop()
                    c_name = line[len("END:") :].strip().upper()
                    if c_name != component.name:
                        raise ValueError(
                            f"Got END:{c_name}, expected END:{component.name}"
                            + f" at line {_i + 1}"
                        )
                    if stack:
                        stack[-1].subcomponents.append(component)
                    else:
                        rv.append(component)
                else:
                    if line.strip():
                        stack[-1].props.append(line)
        except IndexError:
            raise ValueError(f"Parsing error at line {_i + 1}")

        if len(stack) > 0:
            raise ValueError(
                f"Missing END for component(s): {', '.join(c.name for c in stack)}"
            )

        if multiple:
            return rv
        elif len(rv) != 1:
            raise ValueError(f"Found {len(rv)} components, expected one.")
        else:
            return rv[0]

    def dump_lines(self):
        yield f"BEGIN:{self.name}"
        yield from self.props
        for c in self.subcomponents:
            yield from c.dump_lines()
        yield f"END:{self.name}"

    def __delitem__(self, key):
        prefix = (f"{key}:", f"{key};")
        new_lines = []
        lineiter = iter(self.props)
        while True:
            for line in lineiter:
                if line.startswith(prefix):
                    break
                else:
                    new_lines.append(line)
            else:
                break

            for line in lineiter:
                if not line.startswith((" ", "\t", *prefix)):
                    new_lines.append(line)
                    break

        self.props = new_lines

    def __setitem__(self, key, val):
        assert isinstance(val, str)
        assert "\n" not in val
        del self[key]
        line = f"{key}:{val}"
        self.props.append(line)

    def __contains__(self, obj):
        if isinstance(obj, type(self)):
            return obj not in self.subcomponents and not any(
                obj in x for x in self.subcomponents
            )
        elif isinstance(obj, str):
            return self.get(obj, None) is not None
        else:
            raise ValueError(obj)

    def __getitem__(self, key):
        prefix_without_params = f"{key}:"
        prefix_with_params = f"{key};"
        iterlines = iter(self.props)
        for line in iterlines:
            if line.startswith(prefix_without_params):
                rv = line[len(prefix_without_params) :]
                break
            elif line.startswith(prefix_with_params):
                rv = line[len(prefix_with_params) :].split(":", 1)[-1]
                break
        else:
            raise KeyError

        for line in iterlines:
            if line.startswith((" ", "\t")):
                rv += line[1:]
            else:
                break

        return rv

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def __eq__(self, other):
        return (
            isinstance(other, type(self))
            and self.name == other.name
            and self.props == other.props
            and self.subcomponents == other.subcomponents
        )



from __future__ import annotations

import asyncio
import hashlib
import json
import logging
import sys

import aiohttp
import aiostream

from .. import exceptions
from .utils import handle_collection_not_found
from .utils import handle_storage_init_error
from .utils import load_status
from .utils import save_status
from .utils import storage_class_from_config
from .utils import storage_instance_from_config

# Increase whenever upgrade potentially breaks discovery cache and collections
# should be re-discovered
DISCOVERY_CACHE_VERSION = 1

logger = logging.getLogger(__name__)


def _get_collections_cache_key(pair):
    m = hashlib.sha256()
    j = json.dumps(
        [
            DISCOVERY_CACHE_VERSION,
            pair.collections,
            pair.config_a,
            pair.config_b,
        ],
        sort_keys=True,
    )
    m.update(j.encode("utf-8"))
    return m.hexdigest()


async def collections_for_pair(
    status_path,
    pair,
    from_cache=True,
    list_collections=False,
    *,
    connector: aiohttp.TCPConnector,
):
    """Determine all configured collections for a given pair. Takes care of
    shortcut expansion and result caching.

    :param status_path: The path to the status directory.
    :param from_cache: Whether to load from cache (aborting on cache miss) or
        discover and save to cache.

    :returns: iterable of (collection, (a_args, b_args))
    """
    cache_key = _get_collections_cache_key(pair)
    if from_cache:
        rv = load_status(status_path, pair.name, data_type="collections")
        if rv.get("cache_key", None) == cache_key:
            return list(
                _expand_collections_cache(
                    rv["collections"], pair.config_a, pair.config_b
                )
            )
        elif rv:
            raise exceptions.UserError(
                "Detected change in config file, "
                f"please run `vdirsyncer discover {pair.name}`."
            )
        else:
            raise exceptions.UserError(
                f"Please run `vdirsyncer discover {pair.name}`  before synchronization."
            )

    logger.info(f"Discovering collections for pair {pair.name}")

    a_discovered = DiscoverResult(pair.config_a, connector=connector)
    b_discovered = DiscoverResult(pair.config_b, connector=connector)

    if list_collections:
        # TODO: We should gather data and THEN print, so it can be async.
        await _print_collections(
            pair.config_a["instance_name"],
            a_discovered.get_self,
            connector=connector,
        )
        await _print_collections(
            pair.config_b["instance_name"],
            b_discovered.get_self,
            connector=connector,
        )

    # We have to use a list here because the special None/null value would get
    # mangled to string (because JSON objects always have string keys).
    rv = await aiostream.stream.list(
        expand_collections(
            shortcuts=pair.collections,
            config_a=pair.config_a,
            config_b=pair.config_b,
            get_a_discovered=a_discovered.get_self,
            get_b_discovered=b_discovered.get_self,
            _handle_collection_not_found=handle_collection_not_found,
        )
    )

    await _sanity_check_collections(rv, connector=connector)

    save_status(
        base_path=status_path,
        pair=pair.name,
        data_type="collections",
        data={
            "collections": list(
                _compress_collections_cache(rv, pair.config_a, pair.config_b)
            ),
            "cache_key": cache_key,
        },
    )
    return rv


async def _sanity_check_collections(collections, *, connector):
    tasks = []

    for _, (a_args, b_args) in collections:
        tasks.append(storage_instance_from_config(a_args, connector=connector))
        tasks.append(storage_instance_from_config(b_args, connector=connector))

    await asyncio.gather(*tasks)


def _compress_collections_cache(collections, config_a, config_b):
    def deduplicate(x, y):
        rv = {}
        for key, value in x.items():
            if key not in y or y[key] != value:
                rv[key] = value

        return rv

    for name, (a, b) in collections:
        yield name, (deduplicate(a, config_a), deduplicate(b, config_b))


def _expand_collections_cache(collections, config_a, config_b):
    for name, (a_delta, b_delta) in collections:
        a = dict(config_a)
        a.update(a_delta)

        b = dict(config_b)
        b.update(b_delta)

        yield name, (a, b)


class DiscoverResult:
    def __init__(self, config, *, connector):
        self._cls, _ = storage_class_from_config(config)

        if self._cls.__name__ in [
            "CardDAVStorage",
            "CalDAVStorage",
            "GoogleCalendarStorage",
            "GoogleContactsStorage",
        ]:
            assert connector is not None
            config["connector"] = connector

        self._config = config
        self._discovered = None

    async def get_self(self):
        if self._discovered is None:
            self._discovered = await self._discover()
        return self._discovered

    async def _discover(self):
        try:
            discovered = await aiostream.stream.list(self._cls.discover(**self._config))
        except NotImplementedError:
            return {}
        except Exception:
            return handle_storage_init_error(self._cls, self._config)
        else:
            storage_type = self._config["type"]
            rv = {}
            for args in discovered:
                args["type"] = storage_type
                rv[args["collection"]] = args
            return rv


async def expand_collections(
    shortcuts,
    config_a,
    config_b,
    get_a_discovered,
    get_b_discovered,
    _handle_collection_not_found,
):
    handled_collections = set()

    if shortcuts is None:
        shortcuts = [None]

    for shortcut in shortcuts:
        if shortcut == "from a":
            collections = await get_a_discovered()
        elif shortcut == "from b":
            collections = await get_b_discovered()
        else:
            collections = [shortcut]

        for collection in collections:
            if isinstance(collection, list):
                collection, collection_a, collection_b = collection
            else:
                collection_a = collection_b = collection

            if collection in handled_collections:
                continue
            handled_collections.add(collection)

            a_args = await _collection_from_discovered(
                get_a_discovered,
                collection_a,
                config_a,
                _handle_collection_not_found,
            )
            b_args = await _collection_from_discovered(
                get_b_discovered,
                collection_b,
                config_b,
                _handle_collection_not_found,
            )

            yield collection, (a_args, b_args)


async def _collection_from_discovered(
    get_discovered, collection, config, _handle_collection_not_found
):
    if collection is None:
        args = dict(config)
        args["collection"] = None
        return args

    try:
        return (await get_discovered())[collection]
    except KeyError:
        return await _handle_collection_not_found(config, collection)


async def _print_collections(
    instance_name: str,
    get_discovered,
    *,
    connector: aiohttp.TCPConnector,
):
    try:
        discovered = await get_discovered()
    except exceptions.UserError:
        raise
    except Exception:
        # Unless discovery failed due to a user-inflicted error (instanceof
        # UserError), we don't even know if the storage supports discovery
        # properly. So we can't abort.
        import traceback

        logger.debug("".join(traceback.format_tb(sys.exc_info()[2])))
        logger.warning(
            f"Failed to discover collections for {instance_name}, use `-vdebug` "
            "to see the full traceback."
        )
        return
    logger.info(f"{instance_name}:")
    tasks = []
    for args in discovered.values():
        tasks.append(_print_single_collection(args, instance_name, connector))

    await asyncio.gather(*tasks)


async def _print_single_collection(args, instance_name, connector):
    collection = args["collection"]
    if collection is None:
        return

    args["instance_name"] = instance_name
    try:
        storage = await storage_instance_from_config(
            args,
            create=False,
            connector=connector,
        )
        displayname = await storage.get_meta("displayname")
    except Exception:
        displayname = ""

    logger.info(
        "  - {}{}".format(
            json.dumps(collection),
            f' ("{displayname}")' if displayname and displayname != collection else "",
        )
    )



from __future__ import annotations

import abc
import contextlib
import sqlite3
import sys

from .exceptions import IdentAlreadyExists


@contextlib.contextmanager
def _exclusive_transaction(conn):
    c = None
    try:
        c = conn.execute("BEGIN EXCLUSIVE TRANSACTION")
        yield c
        c.execute("COMMIT")
    except BaseException:
        if c is None:
            raise
        _, e, tb = sys.exc_info()
        c.execute("ROLLBACK")
        raise e.with_traceback(tb)


class _StatusBase(metaclass=abc.ABCMeta):
    def load_legacy_status(self, status):
        with self.transaction():
            for ident, metadata in status.items():
                if len(metadata) == 4:
                    href_a, etag_a, href_b, etag_b = metadata
                    props_a = ItemMetadata(href=href_a, hash="UNDEFINED", etag=etag_a)
                    props_b = ItemMetadata(href=href_b, hash="UNDEFINED", etag=etag_b)
                else:
                    a, b = metadata
                    a.setdefault("hash", "UNDEFINED")
                    b.setdefault("hash", "UNDEFINED")
                    props_a = ItemMetadata(**a)
                    props_b = ItemMetadata(**b)

                self.insert_ident_a(ident, props_a)
                self.insert_ident_b(ident, props_b)

    def to_legacy_status(self):
        for ident in self.iter_old():
            a = self.get_a(ident)
            b = self.get_b(ident)
            yield ident, (a.to_status(), b.to_status())

    @abc.abstractmethod
    def transaction(self):
        raise NotImplementedError

    @abc.abstractmethod
    def insert_ident_a(self, ident, props):
        raise NotImplementedError

    @abc.abstractmethod
    def insert_ident_b(self, ident, props):
        raise NotImplementedError

    @abc.abstractmethod
    def update_ident_a(self, ident, props):
        raise NotImplementedError

    @abc.abstractmethod
    def update_ident_b(self, ident, props):
        raise NotImplementedError

    @abc.abstractmethod
    def remove_ident(self, ident):
        raise NotImplementedError

    @abc.abstractmethod
    def get_a(self, ident):
        raise NotImplementedError

    @abc.abstractmethod
    def get_b(self, ident):
        raise NotImplementedError

    @abc.abstractmethod
    def get_new_a(self, ident):
        raise NotImplementedError

    @abc.abstractmethod
    def get_new_b(self, ident):
        raise NotImplementedError

    @abc.abstractmethod
    def iter_old(self):
        raise NotImplementedError

    @abc.abstractmethod
    def iter_new(self):
        raise NotImplementedError

    @abc.abstractmethod
    def get_by_href_a(self, href, default=(None, None)):
        raise NotImplementedError

    @abc.abstractmethod
    def get_by_href_b(self, href, default=(None, None)):
        raise NotImplementedError

    @abc.abstractmethod
    def rollback(self, ident):
        raise NotImplementedError


class SqliteStatus(_StatusBase):
    SCHEMA_VERSION = 1

    def __init__(self, path=":memory:"):
        self._path = path
        self._c = sqlite3.connect(path)
        self._c.isolation_level = None  # turn off idiocy of DB-API
        self._c.row_factory = sqlite3.Row
        self._update_schema()

    def _update_schema(self):
        if self._is_latest_version():
            return

        # If we ever bump the schema version, we will need a way to migrate
        # data.
        with _exclusive_transaction(self._c) as c:
            c.execute('CREATE TABLE meta ( "version" INTEGER PRIMARY KEY )')
            c.execute("INSERT INTO meta (version) VALUES (?)", (self.SCHEMA_VERSION,))

            # I know that this is a bad schema, but right there is just too
            # little gain in deduplicating the .._a and .._b columns.
            c.execute(
                """CREATE TABLE status (
                "ident" TEXT PRIMARY KEY NOT NULL,
                "href_a" TEXT,
                "href_b" TEXT,
                "hash_a" TEXT NOT NULL,
                "hash_b" TEXT NOT NULL,
                "etag_a" TEXT,
                "etag_b" TEXT
            ); """
            )
            c.execute("CREATE UNIQUE INDEX by_href_a ON status(href_a)")
            c.execute("CREATE UNIQUE INDEX by_href_b ON status(href_b)")

            # We cannot add NOT NULL here because data is first fetched for the
            # storage a, then storage b. Inbetween the `_b`-columns are filled
            # with NULL.
            #
            # In an ideal world we would be able to start a transaction with
            # one cursor, write our new data into status and simultaneously
            # query the old status data using a different cursor.
            # Unfortunately sqlite enforces NOT NULL constraints immediately,
            # not just at commit. Since there is also no way to alter
            # constraints on a table (disable constraints on start of
            # transaction and reenable on end), it's a separate table now that
            # just gets copied over before we commit.  That's a lot of copying,
            # sadly.
            c.execute(
                """CREATE TABLE new_status (
                "ident" TEXT PRIMARY KEY NOT NULL,
                "href_a" TEXT,
                "href_b" TEXT,
                "hash_a" TEXT,
                "hash_b" TEXT,
                "etag_a" TEXT,
                "etag_b" TEXT
            ); """
            )

    def close(self):
        if self._c:
            self._c.close()
            self._c = None

    def _is_latest_version(self):
        try:
            return bool(
                self._c.execute(
                    "SELECT version FROM meta WHERE version = ?", (self.SCHEMA_VERSION,)
                ).fetchone()
            )
        except sqlite3.OperationalError:
            return False

    @contextlib.contextmanager
    def transaction(self):
        old_c = self._c
        try:
            with _exclusive_transaction(self._c) as new_c:
                self._c = new_c
                yield
                self._c.execute("DELETE FROM status")
                self._c.execute("INSERT INTO status SELECT * FROM new_status")
                self._c.execute("DELETE FROM new_status")
        finally:
            self._c = old_c

    def insert_ident_a(self, ident, a_props):
        # FIXME: Super inefficient
        old_props = self.get_new_a(ident)
        if old_props is not None:
            raise IdentAlreadyExists(old_href=old_props.href, new_href=a_props.href)
        b_props = self.get_new_b(ident) or ItemMetadata()
        self._c.execute(
            "INSERT OR REPLACE INTO new_status VALUES(?, ?, ?, ?, ?, ?, ?)",
            (
                ident,
                a_props.href,
                b_props.href,
                a_props.hash,
                b_props.hash,
                a_props.etag,
                b_props.etag,
            ),
        )

    def insert_ident_b(self, ident, b_props):
        # FIXME: Super inefficient
        old_props = self.get_new_b(ident)
        if old_props is not None:
            raise IdentAlreadyExists(old_href=old_props.href, new_href=b_props.href)
        a_props = self.get_new_a(ident) or ItemMetadata()
        self._c.execute(
            "INSERT OR REPLACE INTO new_status VALUES(?, ?, ?, ?, ?, ?, ?)",
            (
                ident,
                a_props.href,
                b_props.href,
                a_props.hash,
                b_props.hash,
                a_props.etag,
                b_props.etag,
            ),
        )

    def update_ident_a(self, ident, props):
        self._c.execute(
            "UPDATE new_status SET href_a=?, hash_a=?, etag_a=? WHERE ident=?",
            (props.href, props.hash, props.etag, ident),
        )
        assert self._c.rowcount > 0

    def update_ident_b(self, ident, props):
        self._c.execute(
            "UPDATE new_status SET href_b=?, hash_b=?, etag_b=? WHERE ident=?",
            (props.href, props.hash, props.etag, ident),
        )
        assert self._c.rowcount > 0

    def remove_ident(self, ident):
        self._c.execute("DELETE FROM new_status WHERE ident=?", (ident,))

    def _get_impl(self, ident, side, table):
        res = self._c.execute(
            f"SELECT href_{side} AS href,"
            f"       hash_{side} AS hash,"
            f"       etag_{side} AS etag "
            f"FROM {table} WHERE ident=?",
            (ident,),
        ).fetchone()
        if res is None:
            return None

        if res["hash"] is None:  # FIXME: Implement as constraint in db
            assert res["href"] is None
            assert res["etag"] is None
            return None

        res = dict(res)
        return ItemMetadata(**res)

    def get_a(self, ident):
        return self._get_impl(ident, side="a", table="status")

    def get_b(self, ident):
        return self._get_impl(ident, side="b", table="status")

    def get_new_a(self, ident):
        return self._get_impl(ident, side="a", table="new_status")

    def get_new_b(self, ident):
        return self._get_impl(ident, side="b", table="new_status")

    def iter_old(self):
        return iter(
            res["ident"]
            for res in self._c.execute("SELECT ident FROM status").fetchall()
        )

    def iter_new(self):
        return iter(
            res["ident"]
            for res in self._c.execute("SELECT ident FROM new_status").fetchall()
        )

    def rollback(self, ident):
        a = self.get_a(ident)
        b = self.get_b(ident)
        assert (a is None) == (b is None)

        if a is None and b is None:
            self.remove_ident(ident)
            return

        self._c.execute(
            "INSERT OR REPLACE INTO new_status VALUES (?, ?, ?, ?, ?, ?, ?)",
            (ident, a.href, b.href, a.hash, b.hash, a.etag, b.etag),
        )

    def _get_by_href_impl(self, href, default=(None, None), side=None):
        res = self._c.execute(
            f"SELECT ident, hash_{side} AS hash, etag_{side} AS etag "
            f"FROM status WHERE href_{side}=?",
            (href,),
        ).fetchone()
        if not res:
            return default
        return res["ident"], ItemMetadata(
            href=href,
            hash=res["hash"],
            etag=res["etag"],
        )

    def get_by_href_a(self, *a, **kw):
        kw["side"] = "a"
        return self._get_by_href_impl(*a, **kw)

    def get_by_href_b(self, *a, **kw):
        kw["side"] = "b"
        return self._get_by_href_impl(*a, **kw)


class SubStatus:
    def __init__(self, parent: SqliteStatus, side: str):
        self.parent = parent
        assert side in "ab"

        self.remove_ident = parent.remove_ident

        if side == "a":
            self.insert_ident = parent.insert_ident_a
            self.update_ident = parent.update_ident_a
            self.get = parent.get_a
            self.get_new = parent.get_new_a
            self.get_by_href = parent.get_by_href_a
        else:
            self.insert_ident = parent.insert_ident_b
            self.update_ident = parent.update_ident_b
            self.get = parent.get_b
            self.get_new = parent.get_new_b
            self.get_by_href = parent.get_by_href_b


class ItemMetadata:
    href = None
    hash = None
    etag = None

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            assert hasattr(self, k)
            setattr(self, k, v)

    def to_status(self):
        return {"href": self.href, "etag": self.etag, "hash": self.hash}



from __future__ import annotations

import contextlib
import errno
import importlib
import json
import os
import sys
from typing import Any

import aiohttp
import click

from .. import BUGTRACKER_HOME
from .. import DOCS_HOME
from .. import exceptions
from ..storage.base import Storage
from ..sync.exceptions import IdentConflict
from ..sync.exceptions import PartialSync
from ..sync.exceptions import StorageEmpty
from ..sync.exceptions import SyncConflict
from ..sync.status import SqliteStatus
from ..utils import atomic_write
from ..utils import expand_path
from ..utils import get_storage_init_args
from . import cli_logger

STATUS_PERMISSIONS = 0o600
STATUS_DIR_PERMISSIONS = 0o700


class _StorageIndex:
    def __init__(self):
        self._storages: dict[str, str] = {
            "caldav": "vdirsyncer.storage.dav.CalDAVStorage",
            "carddav": "vdirsyncer.storage.dav.CardDAVStorage",
            "filesystem": "vdirsyncer.storage.filesystem.FilesystemStorage",
            "http": "vdirsyncer.storage.http.HttpStorage",
            "singlefile": "vdirsyncer.storage.singlefile.SingleFileStorage",
            "google_calendar": "vdirsyncer.storage.google.GoogleCalendarStorage",
            "google_contacts": "vdirsyncer.storage.google.GoogleContactsStorage",
        }

    def __getitem__(self, name: str) -> Storage:
        item = self._storages[name]
        if not isinstance(item, str):
            return item

        modname, clsname = item.rsplit(".", 1)
        mod = importlib.import_module(modname)
        self._storages[name] = rv = getattr(mod, clsname)
        assert rv.storage_name == name
        return rv


storage_names = _StorageIndex()
del _StorageIndex


class JobFailed(RuntimeError):
    pass


def handle_cli_error(status_name=None, e=None):
    """
    Print a useful error message for the current exception.

    This is supposed to catch all exceptions, and should never raise any
    exceptions itself.
    """

    try:
        if e is not None:
            raise e
        else:
            raise
    except exceptions.UserError as e:
        cli_logger.critical(e)
    except StorageEmpty as e:
        cli_logger.error(
            '{status_name}: Storage "{name}" was completely emptied. If you '
            "want to delete ALL entries on BOTH sides, then use "
            "`vdirsyncer sync --force-delete {status_name}`. "
            "Otherwise delete the files for {status_name} in your status "
            "directory.".format(
                name=e.empty_storage.instance_name, status_name=status_name
            )
        )
    except PartialSync as e:
        cli_logger.error(
            f"{status_name}: Attempted change on {e.storage}, which is read-only"
            ". Set `partial_sync` in your pair section to `ignore` to ignore "
            "those changes, or `revert` to revert them on the other side."
        )
    except SyncConflict as e:
        cli_logger.error(
            f"{status_name}: One item changed on both sides. Resolve this "
            "conflict manually, or by setting the `conflict_resolution` "
            "parameter in your config file.\n"
            f"See also {DOCS_HOME}/config.html#pair-section\n"
            f"Item ID: {e.ident}\n"
            f"Item href on side A: {e.href_a}\n"
            f"Item href on side B: {e.href_b}\n"
        )
    except IdentConflict as e:
        cli_logger.error(
            '{status_name}: Storage "{storage.instance_name}" contains '
            "multiple items with the same UID or even content. Vdirsyncer "
            "will now abort the synchronization of this collection, because "
            "the fix for this is not clear; It could be the result of a badly "
            "behaving server. You can try running:\n\n"
            "    vdirsyncer repair {storage.instance_name}\n\n"
            "But make sure to have a backup of your data in some form. The "
            "offending hrefs are:\n\n{href_list}\n".format(
                status_name=status_name,
                storage=e.storage,
                href_list="\n".join(map(repr, e.hrefs)),
            )
        )
    except (click.Abort, KeyboardInterrupt, JobFailed):
        pass
    except exceptions.PairNotFound as e:
        cli_logger.error(
            f"Pair {e.pair_name} does not exist. Please check your "
            "configuration file and make sure you've typed the pair name "
            "correctly"
        )
    except exceptions.InvalidResponse as e:
        cli_logger.error(
            "The server returned something vdirsyncer doesn't understand. "
            f"Error message: {e!r}\n"
            "While this is most likely a serverside problem, the vdirsyncer "
            "devs are generally interested in such bugs. Please report it in "
            f"the issue tracker at {BUGTRACKER_HOME}"
        )
    except exceptions.CollectionRequired:
        cli_logger.error(
            "One or more storages don't support `collections = null`. "
            'You probably want to set `collections = ["from a", "from b"]`.'
        )
    except Exception as e:
        tb = sys.exc_info()[2]
        import traceback

        tb = traceback.format_tb(tb)
        if status_name:
            msg = f"Unknown error occurred for {status_name}"
        else:
            msg = "Unknown error occurred"

        msg += f": {e}\nUse `-vdebug` to see the full traceback."

        cli_logger.error(msg)
        cli_logger.debug("".join(tb))


def get_status_name(pair: str, collection: str | None) -> str:
    if collection is None:
        return pair
    return pair + "/" + collection


def get_status_path(
    base_path: str,
    pair: str,
    collection: str | None = None,
    data_type: str | None = None,
) -> str:
    assert data_type is not None
    status_name = get_status_name(pair, collection)
    path = expand_path(os.path.join(base_path, status_name))
    if os.path.isfile(path) and data_type == "items":
        new_path = path + ".items"
        # XXX: Legacy migration
        cli_logger.warning(f"Migrating statuses: Renaming {path} to {new_path}")
        os.rename(path, new_path)

    path += "." + data_type
    return path


def load_status(
    base_path: str,
    pair: str,
    collection: str | None = None,
    data_type: str | None = None,
) -> dict[str, Any]:
    path = get_status_path(base_path, pair, collection, data_type)
    if not os.path.exists(path):
        return {}
    assert_permissions(path, STATUS_PERMISSIONS)

    with open(path) as f:
        try:
            return dict(json.load(f))
        except ValueError:
            pass

    return {}


def prepare_status_path(path: str) -> None:
    dirname = os.path.dirname(path)

    try:
        os.makedirs(dirname, STATUS_DIR_PERMISSIONS)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


@contextlib.contextmanager
def manage_sync_status(base_path: str, pair_name: str, collection_name: str):
    path = get_status_path(base_path, pair_name, collection_name, "items")
    status = None
    legacy_status = None
    try:
        # XXX: Legacy migration
        with open(path, "rb") as f:
            if f.read(1) == b"{":
                f.seek(0)
                legacy_status = dict(json.load(f))
    except (OSError, ValueError):
        pass

    if legacy_status is not None:
        cli_logger.warning("Migrating legacy status to sqlite")
        os.remove(path)
        status = SqliteStatus(path)
        status.load_legacy_status(legacy_status)
    else:
        prepare_status_path(path)
        status = SqliteStatus(path)

    with contextlib.closing(status):
        yield status


def save_status(
    base_path: str,
    pair: str,
    data_type: str,
    data: dict[str, Any],
    collection: str | None = None,
) -> None:
    status_name = get_status_name(pair, collection)
    path = expand_path(os.path.join(base_path, status_name)) + "." + data_type
    prepare_status_path(path)

    with atomic_write(path, mode="w", overwrite=True) as f:
        json.dump(data, f)

    os.chmod(path, STATUS_PERMISSIONS)


def storage_class_from_config(config):
    config = dict(config)
    storage_name = config.pop("type")
    try:
        cls = storage_names[storage_name]
    except KeyError:
        raise exceptions.UserError(f"Unknown storage type: {storage_name}")
    return cls, config


async def storage_instance_from_config(
    config,
    create=True,
    *,
    connector: aiohttp.TCPConnector,
):
    """
    :param config: A configuration dictionary to pass as kwargs to the class
        corresponding to config['type']
    """
    from vdirsyncer.storage.dav import DAVStorage
    from vdirsyncer.storage.http import HttpStorage

    cls, new_config = storage_class_from_config(config)

    if issubclass(cls, DAVStorage) or issubclass(cls, HttpStorage):
        assert connector is not None  # FIXME: hack?
        new_config["connector"] = connector

    try:
        return cls(**new_config)
    except exceptions.CollectionNotFound as e:
        if create:
            config = await handle_collection_not_found(
                config, config.get("collection", None), e=str(e)
            )
            return await storage_instance_from_config(
                config,
                create=False,
                connector=connector,
            )
        else:
            raise
    except Exception:
        return handle_storage_init_error(cls, new_config)


def handle_storage_init_error(cls, config):
    e = sys.exc_info()[1]
    if not isinstance(e, TypeError) or "__init__" not in repr(e):
        raise

    all, required = get_storage_init_args(cls)
    given = set(config)
    missing = required - given
    invalid = given - all

    problems = []

    if missing:
        problems.append(
            "{} storage requires the parameters: {}".format(
                cls.storage_name, ", ".join(missing)
            )
        )

    if invalid:
        problems.append(
            "{} storage doesn't take the parameters: {}".format(
                cls.storage_name, ", ".join(invalid)
            )
        )

    if not problems:
        raise e

    raise exceptions.UserError(
        "Failed to initialize {}".format(config["instance_name"]), problems=problems
    )


def assert_permissions(path: str, wanted: int) -> None:
    permissions = os.stat(path).st_mode & 0o777
    if permissions > wanted:
        cli_logger.warning(
            f"Correcting permissions of {path} from {permissions:o} to {wanted:o}"
        )
        os.chmod(path, wanted)


async def handle_collection_not_found(config, collection, e=None):
    storage_name = config.get("instance_name", None)

    cli_logger.warning(
        "{}No collection {} found for storage {}.".format(
            f"{e}\n" if e else "", json.dumps(collection), storage_name
        )
    )

    if click.confirm("Should vdirsyncer attempt to create it?"):
        storage_type = config["type"]
        cls, config = storage_class_from_config(config)
        config["collection"] = collection
        try:
            args = await cls.create_collection(**config)
            args["type"] = storage_type
            return args
        except NotImplementedError as e:
            cli_logger.error(e)

    raise exceptions.UserError(
        f'Unable to find or create collection "{collection}" for '
        f'storage "{storage_name}". Please create the collection '
        "yourself."
    )



from __future__ import annotations

import logging

from . import exceptions
from .storage.base import normalize_meta_value

logger = logging.getLogger(__name__)


class MetaSyncError(exceptions.Error):
    pass


class MetaSyncConflict(MetaSyncError):
    key = None


def status_set_key(status, key, value):
    if value is None:
        status.pop(key, None)
    else:
        status[key] = value


async def metasync(storage_a, storage_b, status, keys, conflict_resolution=None):
    async def _a_to_b():
        logger.info(f"Copying {key} to {storage_b}")
        await storage_b.set_meta(key, a)
        status_set_key(status, key, a)

    async def _b_to_a():
        logger.info(f"Copying {key} to {storage_a}")
        await storage_a.set_meta(key, b)
        status_set_key(status, key, b)

    async def _resolve_conflict():
        if a == b:
            status_set_key(status, key, a)
        elif conflict_resolution == "a wins":
            await _a_to_b()
        elif conflict_resolution == "b wins":
            await _b_to_a()
        else:
            if callable(conflict_resolution):
                logger.warning("Custom commands don't work on metasync.")
            elif conflict_resolution is not None:
                raise exceptions.UserError("Invalid conflict resolution setting.")
            raise MetaSyncConflict(key)

    for key in keys:
        a = await storage_a.get_meta(key)
        b = await storage_b.get_meta(key)
        s = normalize_meta_value(status.get(key))
        logger.debug(f"Key: {key}")
        logger.debug(f"A: {a}")
        logger.debug(f"B: {b}")
        logger.debug(f"S: {s}")

        if a != s and b != s or storage_a.read_only or storage_b.read_only:
            await _resolve_conflict()
        elif a != s and b == s:
            await _a_to_b()
        elif a == s and b != s:
            await _b_to_a()
        else:
            assert a == b

    for key in set(status) - set(keys):
        del status[key]



from __future__ import annotations

import contextlib
import functools
import os
import sys
import tempfile
import uuid
from inspect import getfullargspec
from typing import Callable

from . import exceptions

# This is only a subset of the chars allowed per the spec. In particular `@` is
# not included, because there are some servers that (incorrectly) encode it to
# `%40` when it's part of a URL path, and reject or "repair" URLs that contain
# `@` in the path. So it's better to just avoid it.
SAFE_UID_CHARS = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_.-+"


_missing = object()


def expand_path(p: str) -> str:
    """Expand $HOME in a path and normalise slashes."""
    p = os.path.expanduser(p)
    p = os.path.normpath(p)
    return p


def split_dict(d: dict, f: Callable):
    """Puts key into first dict if f(key), otherwise in second dict"""
    a = {}
    b = {}
    for k, v in d.items():
        if f(k):
            a[k] = v
        else:
            b[k] = v
    return a, b


def uniq(s):
    """Filter duplicates while preserving order. ``set`` can almost always be
    used instead of this, but preserving order might prove useful for
    debugging."""
    d = set()
    for x in s:
        if x not in d:
            d.add(x)
            yield x


def get_etag_from_file(f):
    """Get etag from a filepath or file-like object.

    This function will flush/sync the file as much as necessary to obtain a
    correct value.
    """
    if hasattr(f, "read"):
        f.flush()  # Only this is necessary on Linux
        if sys.platform == "win32":
            os.fsync(f.fileno())  # Apparently necessary on Windows
        stat = os.fstat(f.fileno())
    else:
        stat = os.stat(f)

    mtime = getattr(stat, "st_mtime_ns", None)
    if mtime is None:
        mtime = stat.st_mtime
    return f"{mtime:.9f};{stat.st_ino}"


def get_storage_init_specs(cls, stop_at=object):
    if cls is stop_at:
        return ()

    spec = getfullargspec(cls.__init__)
    traverse_superclass = getattr(cls.__init__, "_traverse_superclass", True)
    if traverse_superclass:
        if traverse_superclass is True:  # noqa
            supercls = next(
                getattr(x.__init__, "__objclass__", x) for x in cls.__mro__[1:]
            )
        else:
            supercls = traverse_superclass
        superspecs = get_storage_init_specs(supercls, stop_at=stop_at)
    else:
        superspecs = ()

    return (spec,) + superspecs


def get_storage_init_args(cls, stop_at=object):
    """
    Get args which are taken during class initialization. Assumes that all
    classes' __init__ calls super().__init__ with the rest of the arguments.

    :param cls: The class to inspect.
    :returns: (all, required), where ``all`` is a set of all arguments the
        class can take, and ``required`` is the subset of arguments the class
        requires.
    """
    all, required = set(), set()
    for spec in get_storage_init_specs(cls, stop_at=stop_at):
        all.update(spec.args[1:])
        last = -len(spec.defaults) if spec.defaults else len(spec.args)
        required.update(spec.args[1:last])

    return all, required


def checkdir(path: str, create: bool = False, mode: int = 0o750) -> None:
    """Check whether ``path`` is a directory.

    :param create: Whether to create the directory (and all parent directories)
        if it does not exist.
    :param mode: Mode to create missing directories with.
    """

    if not os.path.isdir(path):
        if os.path.exists(path):
            raise OSError(f"{path} is not a directory.")
        if create:
            os.makedirs(path, mode)
        else:
            raise exceptions.CollectionNotFound(f"Directory {path} does not exist.")


def checkfile(path, create=False) -> None:
    """Check whether ``path`` is a file.

    :param create: Whether to create the file's parent directories if they do
        not exist.
    :raises CollectionNotFound: if path does not exist.
    :raises OSError: if path exists but is not a file.
    """
    checkdir(os.path.dirname(path), create=create)
    if not os.path.isfile(path):
        if os.path.exists(path):
            raise OSError(f"{path} is not a file.")
        if create:
            with open(path, "wb"):
                pass
        else:
            raise exceptions.CollectionNotFound(f"File {path} does not exist.")


class cached_property:
    """A read-only @property that is only evaluated once. Only usable on class
    instances' methods.
    """

    def __init__(self, fget, doc=None):
        self.__name__ = fget.__name__
        self.__module__ = fget.__module__
        self.__doc__ = doc or fget.__doc__
        self.fget = fget

    def __get__(self, obj, cls):
        if obj is None:  # pragma: no cover
            return self
        obj.__dict__[self.__name__] = result = self.fget(obj)
        return result


def href_safe(ident, safe=SAFE_UID_CHARS):
    return not bool(set(ident) - set(safe))


def generate_href(ident=None, safe=SAFE_UID_CHARS):
    """
    Generate a safe identifier, suitable for URLs, storage hrefs or UIDs.

    If the given ident string is safe, it will be returned, otherwise a random
    UUID.
    """
    if not ident or not href_safe(ident, safe):
        return str(uuid.uuid4())
    else:
        return ident


def synchronized(lock=None):
    if lock is None:
        from threading import Lock

        lock = Lock()

    def inner(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            with lock:
                return f(*args, **kwargs)

        return wrapper

    return inner


def open_graphical_browser(url, new=0, autoraise=True):
    """Open a graphical web browser.

    This is basically like `webbrowser.open`, but without trying to launch CLI
    browsers at all. We're excluding those since it's undesirable to launch
    those when you're using vdirsyncer on a server. Rather copypaste the URL
    into the local browser, or use the URL-yanking features of your terminal
    emulator.
    """
    import webbrowser

    cli_names = {"www-browser", "links", "links2", "elinks", "lynx", "w3m"}

    if webbrowser._tryorder is None:  # Python 3.8
        webbrowser.register_standard_browsers()

    for name in webbrowser._tryorder:
        if name in cli_names:
            continue

        browser = webbrowser.get(name)
        if browser.open(url, new, autoraise):
            return

    raise RuntimeError("No graphical browser found. Please open the URL manually.")


@contextlib.contextmanager
def atomic_write(dest, mode="wb", overwrite=False):
    if "w" not in mode:
        raise RuntimeError("`atomic_write` requires write access")

    fd, src = tempfile.mkstemp(prefix=os.path.basename(dest), dir=os.path.dirname(dest))
    file = os.fdopen(fd, mode=mode)

    try:
        yield file
    except Exception:
        os.unlink(src)
        raise
    else:
        file.flush()
        file.close()

        if overwrite:
            os.rename(src, dest)
        else:
            os.link(src, dest)
            os.unlink(src)



from __future__ import annotations

import logging
import subprocess
import urllib.parse as urlparse

import aiohttp

from .. import exceptions
from ..http import USERAGENT
from ..http import prepare_auth
from ..http import prepare_client_cert
from ..http import prepare_verify
from ..http import request
from ..vobject import Item
from ..vobject import split_collection
from .base import Storage

logger = logging.getLogger(__name__)


class HttpStorage(Storage):
    storage_name = "http"
    read_only = True
    _repr_attributes = ["username", "url"]
    _items = None

    # Required for tests.
    _ignore_uids = True

    def __init__(
        self,
        url,
        username="",
        password="",
        verify=None,
        auth=None,
        useragent=USERAGENT,
        verify_fingerprint=None,
        auth_cert=None,
        filter_hook=None,
        *,
        connector,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self._settings = {
            "cert": prepare_client_cert(auth_cert),
            "latin1_fallback": False,
        }
        auth = prepare_auth(auth, username, password)
        if auth:
            self._settings["auth"] = auth

        ssl = prepare_verify(verify, verify_fingerprint)
        if ssl:
            self._settings["ssl"] = ssl

        self.username, self.password = username, password
        self.useragent = useragent
        assert connector is not None
        self.connector = connector
        self._filter_hook = filter_hook

        collection = kwargs.get("collection")
        if collection is not None:
            url = urlparse.urljoin(url, collection)
        self.url = url
        self.parsed_url = urlparse.urlparse(self.url)

    def _default_headers(self):
        return {"User-Agent": self.useragent}

    def _run_filter_hook(self, raw_item):
        try:
            result = subprocess.run(
                [self._filter_hook],
                input=raw_item,
                capture_output=True,
                encoding="utf-8",
            )
            return result.stdout
        except OSError as e:
            logger.warning(f"Error executing external command: {str(e)}")
            return raw_item

    async def list(self):
        async with aiohttp.ClientSession(
            connector=self.connector,
            connector_owner=False,
            trust_env=True,
            # TODO use `raise_for_status=true`, though this needs traces first,
        ) as session:
            r = await request(
                "GET",
                self.url,
                headers=self._default_headers(),
                session=session,
                **self._settings,
            )
        self._items = {}

        for raw_item in split_collection((await r.read()).decode("utf-8")):
            if self._filter_hook:
                raw_item = self._run_filter_hook(raw_item)
            if not raw_item:
                continue

            item = Item(raw_item)
            if self._ignore_uids:
                item = item.with_uid(item.hash)

            self._items[item.ident] = item, item.hash

        for href, (_, etag) in self._items.items():
            yield href, etag

    async def get(self, href):
        if self._items is None:
            async for _ in self.list():
                pass

        try:
            return self._items[href]
        except KeyError:
            raise exceptions.NotFoundError(href)



from __future__ import annotations

import datetime
import logging
import urllib.parse as urlparse
import xml.etree.ElementTree as etree
from abc import abstractmethod
from inspect import getfullargspec
from inspect import signature

import aiohttp
import aiostream

from vdirsyncer.exceptions import Error
from vdirsyncer.vobject import Item

from .. import exceptions
from .. import http
from .. import utils
from ..http import USERAGENT
from ..http import prepare_auth
from ..http import prepare_client_cert
from ..http import prepare_verify
from .base import Storage
from .base import normalize_meta_value

dav_logger = logging.getLogger(__name__)

CALDAV_DT_FORMAT = "%Y%m%dT%H%M%SZ"


async def _assert_multistatus_success(r):
    # Xandikos returns a multistatus on PUT.
    try:
        root = _parse_xml(await r.content.read())
    except InvalidXMLResponse:
        return
    for status in root.findall(".//{DAV:}status"):
        parts = status.text.strip().split()
        try:
            st = int(parts[1])
        except (ValueError, IndexError):
            continue
        if st < 200 or st >= 400:
            raise Error(f"Server error: {st}")


def _normalize_href(base, href):
    """Normalize the href to be a path only relative to hostname and schema."""
    orig_href = href
    if not href:
        raise ValueError(href)

    x = urlparse.urljoin(base, href)
    x = urlparse.urlsplit(x).path

    # We unquote and quote again, but want to make sure we
    # keep around the "@" character.
    x = urlparse.unquote(x)
    x = urlparse.quote(x, "/@")

    if orig_href == x:
        dav_logger.debug(f"Already normalized: {x!r}")
    else:
        dav_logger.debug(f"Normalized URL from {orig_href!r} to {x!r}")

    return x


class InvalidXMLResponse(exceptions.InvalidResponse):
    pass


_BAD_XML_CHARS = (
    b"\x00\x01\x02\x03\x04\x05\x06\x07\x08\x0b\x0c\x0e\x0f"
    b"\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f"
)


def _clean_body(content, bad_chars=_BAD_XML_CHARS):
    new_content = content.translate(None, bad_chars)
    if new_content != content:
        dav_logger.warning(
            "Your server incorrectly returned ASCII control characters in its "
            "XML. Vdirsyncer ignores those, but this is a bug in your server."
        )
    return new_content


def _parse_xml(content):
    try:
        return etree.XML(_clean_body(content))
    except etree.ParseError as e:
        raise InvalidXMLResponse(
            f"Invalid XML encountered: {e}\nDouble-check the URLs in your config."
        )


def _merge_xml(items):
    if not items:
        return None
    rv = items[0]
    for item in items[1:]:
        rv.extend(item.iter())
    return rv


def _fuzzy_matches_mimetype(strict, weak):
    # different servers give different getcontenttypes:
    # "text/vcard", "text/x-vcard", "text/x-vcard; charset=utf-8",
    # "text/directory;profile=vCard", "text/directory",
    # "text/vcard; charset=utf-8"
    if strict is None or weak is None:
        return True

    mediatype, subtype = strict.split("/")
    if subtype in weak:
        return True
    return False


class Discover:
    @property
    @abstractmethod
    def _namespace(self) -> str:
        pass

    @property
    @abstractmethod
    def _resourcetype(self) -> str | None:
        pass

    @property
    @abstractmethod
    def _homeset_xml(self) -> bytes:
        pass

    @property
    @abstractmethod
    def _homeset_tag(self) -> str:
        pass

    @property
    @abstractmethod
    def _well_known_uri(self) -> str:
        pass

    _collection_xml = b"""
    <propfind xmlns="DAV:">
        <prop>
            <resourcetype />
        </prop>
    </propfind>
    """

    def __init__(self, session, kwargs):
        if kwargs.pop("collection", None) is not None:
            raise TypeError("collection argument must not be given.")

        self.session = session
        self.kwargs = kwargs

    @staticmethod
    def _get_collection_from_url(url):
        _, collection = url.rstrip("/").rsplit("/", 1)
        return urlparse.unquote(collection)

    async def find_principal(self):
        try:
            return await self._find_principal_impl("")
        except (aiohttp.ClientResponseError, exceptions.Error):
            dav_logger.debug("Trying out well-known URI")
            return await self._find_principal_impl(self._well_known_uri)

    async def _find_principal_impl(self, url) -> str:
        headers = self.session.get_default_headers()
        headers["Depth"] = "0"
        body = b"""
        <propfind xmlns="DAV:">
            <prop>
                <current-user-principal />
            </prop>
        </propfind>
        """

        response = await self.session.request(
            "PROPFIND",
            url,
            headers=headers,
            data=body,
        )

        root = _parse_xml(await response.content.read())
        rv = root.find(".//{DAV:}current-user-principal/{DAV:}href")
        if rv is None:
            # This is for servers that don't support current-user-principal
            # E.g. Synology NAS
            # See https://github.com/pimutils/vdirsyncer/issues/498
            dav_logger.debug(
                f"No current-user-principal returned, re-using URL {response.url}"
            )
            return response.url.human_repr()
        return urlparse.urljoin(str(response.url), rv.text).rstrip("/") + "/"

    async def find_home(self):
        url = await self.find_principal()
        headers = self.session.get_default_headers()
        headers["Depth"] = "0"
        response = await self.session.request(
            "PROPFIND", url, headers=headers, data=self._homeset_xml
        )

        root = etree.fromstring(await response.content.read())
        # Better don't do string formatting here, because of XML namespaces
        rv = root.find(".//" + self._homeset_tag + "/{DAV:}href")
        if rv is None:
            raise InvalidXMLResponse("Couldn't find home-set.")
        return urlparse.urljoin(str(response.url), rv.text).rstrip("/") + "/"

    async def find_collections(self):
        rv = None
        try:
            rv = await aiostream.stream.list(self._find_collections_impl(""))
        except (aiohttp.ClientResponseError, exceptions.Error):
            pass

        if rv:
            return rv

        dav_logger.debug("Given URL is not a homeset URL")
        return await aiostream.stream.list(
            self._find_collections_impl(await self.find_home())
        )

    def _check_collection_resource_type(self, response):
        if self._resourcetype is None:
            return True

        props = _merge_xml(response.findall("{DAV:}propstat/{DAV:}prop"))
        if props is None or not len(props):
            dav_logger.debug("Skipping, missing <prop>: %s", response)
            return False
        if props.find("{DAV:}resourcetype/" + self._resourcetype) is None:
            dav_logger.debug(
                "Skipping, not of resource type %s: %s", self._resourcetype, response
            )
            return False
        return True

    async def _find_collections_impl(self, url):
        headers = self.session.get_default_headers()
        headers["Depth"] = "1"
        r = await self.session.request(
            "PROPFIND", url, headers=headers, data=self._collection_xml
        )
        root = _parse_xml(await r.content.read())
        done = set()
        for response in root.findall("{DAV:}response"):
            if not self._check_collection_resource_type(response):
                continue

            href = response.find("{DAV:}href")
            if href is None:
                raise InvalidXMLResponse("Missing href tag for collection props.")
            href = urlparse.urljoin(str(r.url), href.text)
            if href not in done:
                done.add(href)
                yield {"href": href}

    async def discover(self):
        for c in await self.find_collections():
            url = c["href"]
            collection = self._get_collection_from_url(url)
            storage_args = dict(self.kwargs)
            storage_args.update({"url": url, "collection": collection})
            yield storage_args

    async def create(self, collection):
        if collection is None:
            collection = self._get_collection_from_url(self.kwargs["url"])

        async for c in self.discover():
            if c["collection"] == collection:
                return c

        home = await self.find_home()
        url = urlparse.urljoin(home, urlparse.quote(collection, "/@"))

        try:
            url = await self._create_collection_impl(url)
        except (aiohttp.ClientResponseError, Error) as e:
            raise NotImplementedError(e)
        else:
            rv = dict(self.kwargs)
            rv["collection"] = collection
            rv["url"] = url
            return rv

    async def _create_collection_impl(self, url):
        data = """<?xml version="1.0" encoding="utf-8" ?>
            <mkcol xmlns="DAV:">
                <set>
                    <prop>
                        <resourcetype>
                            <collection/>
                            {}
                        </resourcetype>
                    </prop>
                </set>
            </mkcol>
        """.format(
            etree.tostring(etree.Element(self._resourcetype), encoding="unicode")
        ).encode("utf-8")

        response = await self.session.request(
            "MKCOL",
            url,
            data=data,
            headers=self.session.get_default_headers(),
        )
        return str(response.url)


class CalDiscover(Discover):
    _namespace = "urn:ietf:params:xml:ns:caldav"
    _resourcetype = "{%s}calendar" % _namespace
    _homeset_xml = b"""
    <propfind xmlns="DAV:" xmlns:c="urn:ietf:params:xml:ns:caldav">
        <prop>
            <c:calendar-home-set />
        </prop>
    </propfind>
    """
    _homeset_tag = "{%s}calendar-home-set" % _namespace
    _well_known_uri = "/.well-known/caldav"


class CardDiscover(Discover):
    _namespace = "urn:ietf:params:xml:ns:carddav"
    _resourcetype: str | None = "{%s}addressbook" % _namespace
    _homeset_xml = b"""
    <propfind xmlns="DAV:" xmlns:c="urn:ietf:params:xml:ns:carddav">
        <prop>
            <c:addressbook-home-set />
        </prop>
    </propfind>
    """
    _homeset_tag = "{%s}addressbook-home-set" % _namespace
    _well_known_uri = "/.well-known/carddav"


class DAVSession:
    """A helper class to connect to DAV servers."""

    connector: aiohttp.BaseConnector

    @classmethod
    def init_and_remaining_args(cls, **kwargs):
        def is_arg(k):
            """Return true if ``k`` is an argument of ``cls.__init__``."""
            return k in argspec.args or k in argspec.kwonlyargs

        argspec = getfullargspec(cls.__init__)
        self_args, remainder = utils.split_dict(kwargs, is_arg)

        return cls(**self_args), remainder

    def __init__(
        self,
        url,
        username="",
        password="",
        verify=None,
        auth=None,
        useragent=USERAGENT,
        verify_fingerprint=None,
        auth_cert=None,
        *,
        connector: aiohttp.BaseConnector,
    ):
        self._settings = {
            "cert": prepare_client_cert(auth_cert),
        }
        auth = prepare_auth(auth, username, password)
        if auth:
            self._settings["auth"] = auth

        ssl = prepare_verify(verify, verify_fingerprint)
        if ssl:
            self._settings["ssl"] = ssl

        self.useragent = useragent
        self.url = url.rstrip("/") + "/"
        self.connector = connector

    @utils.cached_property
    def parsed_url(self):
        return urlparse.urlparse(self.url)

    async def request(self, method, path, **kwargs):
        url = self.url
        if path:
            url = urlparse.urljoin(str(self.url), path)

        more = dict(self._settings)
        more.update(kwargs)

        # XXX: This is a temporary hack to pin-point bad refactoring.
        assert self.connector is not None
        async with self._session as session:
            return await http.request(method, url, session=session, **more)

    @property
    def _session(self):
        """Return a new session for requests."""

        return aiohttp.ClientSession(
            connector=self.connector,
            connector_owner=False,
            trust_env=True,
            # TODO use `raise_for_status=true`, though this needs traces first,
        )

    def get_default_headers(self):
        return {
            "User-Agent": self.useragent,
            "Content-Type": "application/xml; charset=UTF-8",
        }


class DAVStorage(Storage):
    # the file extension of items. Useful for testing against radicale.
    fileext: str
    # mimetype of items
    item_mimetype: str

    @property
    @abstractmethod
    def get_multi_template(self) -> str:
        """XML to use when fetching multiple hrefs."""

    @property
    @abstractmethod
    def get_multi_data_query(self) -> str:
        """LXML query for extracting results in get_multi."""

    @property
    @abstractmethod
    def discovery_class(self) -> type[Discover]:
        """Discover subclass to use."""

    # The DAVSession class to use
    session_class = DAVSession

    connector: aiohttp.TCPConnector

    _repr_attributes = ["username", "url"]

    _property_table = {
        "displayname": ("displayname", "DAV:"),
    }

    def __init__(self, *, connector, **kwargs):
        # defined for _repr_attributes
        self.username = kwargs.get("username")
        self.url = kwargs.get("url")
        self.connector = connector

        self.session, kwargs = self.session_class.init_and_remaining_args(
            connector=connector,
            **kwargs,
        )
        super().__init__(**kwargs)

    __init__.__signature__ = signature(session_class.__init__)  # type: ignore
    # See  https://github.com/python/mypy/issues/5958

    @classmethod
    async def discover(cls, **kwargs):
        session, _ = cls.session_class.init_and_remaining_args(**kwargs)
        d = cls.discovery_class(session, kwargs)

        async for collection in d.discover():
            yield collection

    @classmethod
    async def create_collection(cls, collection, **kwargs):
        session, _ = cls.session_class.init_and_remaining_args(**kwargs)
        d = cls.discovery_class(session, kwargs)
        return await d.create(collection)

    def _normalize_href(self, *args, **kwargs):
        return _normalize_href(self.session.url, *args, **kwargs)

    def _get_href(self, item):
        href = utils.generate_href(item.ident)
        return self._normalize_href(href + self.fileext)

    def _is_item_mimetype(self, mimetype):
        return _fuzzy_matches_mimetype(self.item_mimetype, mimetype)

    async def get(self, href: str):
        ((actual_href, item, etag),) = await aiostream.stream.list(
            self.get_multi([href])
        )
        assert href == actual_href
        return item, etag

    async def get_multi(self, hrefs):
        hrefs = set(hrefs)
        href_xml = []
        for href in hrefs:
            if href != self._normalize_href(href):
                raise exceptions.NotFoundError(href)
            href_xml.append(f"<href>{href}</href>")
        if href_xml:
            data = self.get_multi_template.format(hrefs="\n".join(href_xml)).encode(
                "utf-8"
            )
            response = await self.session.request(
                "REPORT", "", data=data, headers=self.session.get_default_headers()
            )
            root = _parse_xml(
                await response.content.read()
            )  # etree only can handle bytes
            rv = []
            hrefs_left = set(hrefs)
            for href, etag, prop in self._parse_prop_responses(root):
                raw = prop.find(self.get_multi_data_query)
                if raw is None:
                    dav_logger.warning(f"Skipping {href}, the item content is missing.")
                    continue

                raw = raw.text or ""

                if isinstance(raw, bytes):
                    raw = raw.decode(response.encoding)
                if isinstance(etag, bytes):
                    etag = etag.decode(response.encoding)

                try:
                    hrefs_left.remove(href)
                except KeyError:
                    if href in hrefs:
                        dav_logger.warning(f"Server sent item twice: {href}")
                    else:
                        dav_logger.warning(f"Server sent unsolicited item: {href}")
                else:
                    rv.append((href, Item(raw), etag))
            for href in hrefs_left:
                raise exceptions.NotFoundError(href)

            for href, item, etag in rv:
                yield href, item, etag

    async def _put(self, href, item, etag):
        headers = self.session.get_default_headers()
        headers["Content-Type"] = self.item_mimetype
        if etag is None:
            headers["If-None-Match"] = "*"
        else:
            headers["If-Match"] = etag

        response = await self.session.request(
            "PUT", href, data=item.raw.encode("utf-8"), headers=headers
        )

        await _assert_multistatus_success(response)

        # The server may not return an etag under certain conditions:
        #
        #   An origin server MUST NOT send a validator header field (Section
        #   7.2), such as an ETag or Last-Modified field, in a successful
        #   response to PUT unless the request's representation data was saved
        #   without any transformation applied to the body (i.e., the
        #   resource's new representation data is identical to the
        #   representation data received in the PUT request) and the validator
        #   field value reflects the new representation.
        #
        # -- https://tools.ietf.org/html/rfc7231#section-4.3.4
        #
        # In such cases we return a constant etag. The next synchronization
        # will then detect an etag change and will download the new item.
        etag = response.headers.get("etag", None)
        href = self._normalize_href(str(response.url))
        return href, etag

    async def update(self, href, item, etag):
        if etag is None:
            raise ValueError("etag must be given and must not be None.")
        href, etag = await self._put(self._normalize_href(href), item, etag)
        return etag

    async def upload(self, item: Item):
        href = self._get_href(item)
        rv = await self._put(href, item, None)
        return rv

    async def delete(self, href, etag):
        href = self._normalize_href(href)
        headers = self.session.get_default_headers()
        if etag:  # baikal doesn't give us an etag.
            dav_logger.warning("Deleting an item with no etag.")
            headers.update({"If-Match": etag})

        await self.session.request("DELETE", href, headers=headers)

    def _parse_prop_responses(self, root, handled_hrefs=None):
        if handled_hrefs is None:
            handled_hrefs = set()
        for response in root.iter("{DAV:}response"):
            href = response.find("{DAV:}href")
            if href is None:
                dav_logger.error("Skipping response, href is missing.")
                continue

            href = self._normalize_href(href.text)

            if href in handled_hrefs:
                # Servers that send duplicate hrefs:
                # - Zimbra
                #   https://github.com/pimutils/vdirsyncer/issues/88
                # - Davmail
                #   https://github.com/pimutils/vdirsyncer/issues/144
                dav_logger.warning(f"Skipping identical href: {href!r}")
                continue

            props = response.findall("{DAV:}propstat/{DAV:}prop")
            if props is None or not len(props):
                dav_logger.debug(f"Skipping {href!r}, properties are missing.")
                continue
            else:
                props = _merge_xml(props)

            if props.find("{DAV:}resourcetype/{DAV:}collection") is not None:
                dav_logger.debug(f"Skipping {href!r}, is collection.")
                continue

            etag = getattr(props.find("{DAV:}getetag"), "text", "")
            if not etag:
                dav_logger.debug(f"Skipping {href!r}, etag property is missing.")
                continue

            contenttype = getattr(props.find("{DAV:}getcontenttype"), "text", None)
            if not self._is_item_mimetype(contenttype):
                dav_logger.debug(
                    f"Skipping {href!r}, {contenttype!r} != {self.item_mimetype!r}."
                )
                continue

            handled_hrefs.add(href)
            yield href, etag, props

    async def list(self):
        headers = self.session.get_default_headers()
        headers["Depth"] = "1"

        data = b"""<?xml version="1.0" encoding="utf-8" ?>
            <propfind xmlns="DAV:">
                <prop>
                    <resourcetype/>
                    <getcontenttype/>
                    <getetag/>
                </prop>
            </propfind>
            """

        # We use a PROPFIND request instead of addressbook-query due to issues
        # with Zimbra. See https://github.com/pimutils/vdirsyncer/issues/83
        response = await self.session.request(
            "PROPFIND",
            "",
            data=data,
            headers=headers,
        )
        root = _parse_xml(await response.content.read())

        rv = self._parse_prop_responses(root)
        for href, etag, _prop in rv:
            yield href, etag

    async def get_meta(self, key) -> str | None:
        try:
            tagname, namespace = self._property_table[key]
        except KeyError:
            raise exceptions.UnsupportedMetadataError

        xpath = f"{{{namespace}}}{tagname}"
        body = f"""<?xml version="1.0" encoding="utf-8" ?>
            <propfind xmlns="DAV:">
                <prop>
                    {etree.tostring(etree.Element(xpath), encoding="unicode")}
                </prop>
            </propfind>
        """
        data = body.encode("utf-8")

        headers = self.session.get_default_headers()
        headers["Depth"] = "0"

        response = await self.session.request(
            "PROPFIND",
            "",
            data=data,
            headers=headers,
        )

        root = _parse_xml(await response.content.read())

        for prop in root.findall(".//" + xpath):
            text = normalize_meta_value(getattr(prop, "text", None))
            if text:
                return text
        return None

    async def set_meta(self, key, value):
        try:
            tagname, namespace = self._property_table[key]
        except KeyError:
            raise exceptions.UnsupportedMetadataError

        lxml_selector = f"{{{namespace}}}{tagname}"
        element = etree.Element(lxml_selector)
        if value is None:
            action = "remove"
        else:
            element.text = normalize_meta_value(value)
            action = "set"

        data = """<?xml version="1.0" encoding="utf-8" ?>
            <propertyupdate xmlns="DAV:">
                <{action}>
                    <prop>
                        {}
                    </prop>
                </{action}>
            </propertyupdate>
        """.format(
            etree.tostring(element, encoding="unicode"),
            action=action,
        ).encode("utf-8")

        await self.session.request(
            "PROPPATCH",
            "",
            data=data,
            headers=self.session.get_default_headers(),
        )

        # XXX: Response content is currently ignored. Though exceptions are
        # raised for HTTP errors, a multistatus with errorcodes inside is not
        # parsed yet. Not sure how common those are, or how they look like. It
        # might be easier (and safer in case of a stupid server) to just issue
        # a PROPFIND to see if the value got actually set.


class CalDAVStorage(DAVStorage):
    storage_name = "caldav"
    fileext = ".ics"
    item_mimetype = "text/calendar"
    discovery_class = CalDiscover

    start_date = None
    end_date = None

    get_multi_template = """<?xml version="1.0" encoding="utf-8" ?>
        <C:calendar-multiget xmlns="DAV:"
            xmlns:C="urn:ietf:params:xml:ns:caldav">
            <prop>
                <getetag/>
                <C:calendar-data/>
            </prop>
            {hrefs}
        </C:calendar-multiget>"""

    get_multi_data_query = "{urn:ietf:params:xml:ns:caldav}calendar-data"

    _property_table = dict(DAVStorage._property_table)
    _property_table.update(
        {
            "color": ("calendar-color", "http://apple.com/ns/ical/"),
            "description": ("calendar-description", "urn:ietf:params:xml:ns:caldav"),
            "order": ("calendar-order", "http://apple.com/ns/ical/"),
        }
    )

    def __init__(self, start_date=None, end_date=None, item_types=(), **kwargs):
        super().__init__(**kwargs)
        if not isinstance(item_types, (list, tuple)):
            raise exceptions.UserError("item_types must be a list.")

        self.item_types = tuple(item_types)
        if (start_date is None) != (end_date is None):
            raise exceptions.UserError(
                "If start_date is given, end_date has to be given too."
            )
        elif start_date is not None and end_date is not None:
            namespace = dict(datetime.__dict__)
            namespace["start_date"] = self.start_date = (
                eval(start_date, namespace)
                if isinstance(start_date, (bytes, str))
                else start_date
            )
            self.end_date = (
                eval(end_date, namespace)
                if isinstance(end_date, (bytes, str))
                else end_date
            )

    @staticmethod
    def _get_list_filters(components, start, end):
        if components:
            caldavfilter = """
                <C:comp-filter name="VCALENDAR">
                    <C:comp-filter name="{component}">
                        {timefilter}
                    </C:comp-filter>
                </C:comp-filter>
                """

            if start is not None and end is not None:
                start = start.strftime(CALDAV_DT_FORMAT)
                end = end.strftime(CALDAV_DT_FORMAT)

                timefilter = f'<C:time-range start="{start}" end="{end}"/>'
            else:
                timefilter = ""

            for component in components:
                yield caldavfilter.format(component=component, timefilter=timefilter)
        else:
            if start is not None and end is not None:
                yield from CalDAVStorage._get_list_filters(
                    ("VTODO", "VEVENT"), start, end
                )

    async def list(self):
        caldavfilters = list(
            self._get_list_filters(self.item_types, self.start_date, self.end_date)
        )
        if not caldavfilters:
            # If we don't have any filters (which is the default), taking the
            # risk of sending a calendar-query is not necessary. There doesn't
            # seem to be a widely-usable way to send calendar-queries with the
            # same semantics as a PROPFIND request... so why not use PROPFIND
            # instead?
            #
            # See https://github.com/dmfs/tasks/issues/118 for backstory.
            async for href, etag in super().list():
                yield href, etag

        data = """<?xml version="1.0" encoding="utf-8" ?>
            <C:calendar-query xmlns="DAV:"
                xmlns:C="urn:ietf:params:xml:ns:caldav">
                <prop>
                    <getcontenttype/>
                    <getetag/>
                </prop>
                <C:filter>
                {caldavfilter}
                </C:filter>
            </C:calendar-query>"""

        headers = self.session.get_default_headers()
        # https://github.com/pimutils/vdirsyncer/issues/166
        # The default in CalDAV's calendar-queries is 0, but the examples use
        # an explicit value of 1 for querying items. it is extremely unclear in
        # the spec which values from WebDAV are actually allowed.
        headers["Depth"] = "1"

        handled_hrefs = set()

        for caldavfilter in caldavfilters:
            xml = data.format(caldavfilter=caldavfilter).encode("utf-8")
            response = await self.session.request(
                "REPORT",
                "",
                data=xml,
                headers=headers,
            )
            root = _parse_xml(await response.content.read())
            rv = self._parse_prop_responses(root, handled_hrefs)
            for href, etag, _prop in rv:
                yield href, etag


class CardDAVStorage(DAVStorage):
    storage_name = "carddav"
    fileext = ".vcf"
    item_mimetype = "text/vcard"
    discovery_class = CardDiscover

    def __init__(self, *args, use_vcard_4=False, **kwargs):
        self.use_vcard_4 = use_vcard_4
        super().__init__(*args, **kwargs)

    @property
    def get_multi_template(self):
        ct = 'Content-Type="text/vcard" version="4.0"' if self.use_vcard_4 else ""
        return f"""<?xml version="1.0" encoding="utf-8" ?>
            <C:addressbook-multiget xmlns="DAV:"
                    xmlns:C="urn:ietf:params:xml:ns:carddav">
                <prop>
                    <getetag/>
                    <C:address-data {ct}/>
                </prop>
                {{hrefs}}
            </C:addressbook-multiget>"""

    get_multi_data_query = "{urn:ietf:params:xml:ns:carddav}address-data"

    _property_table = dict(DAVStorage._property_table)
    _property_table.update(
        {
            "description": (
                "addressbook-description",
                "urn:ietf:params:xml:ns:carddav",
            ),
        }
    )
